# coding=utf-8
import os
import c4d
import math
import random
from c4d import plugins, bitmaps, gui, documents, Vector
from c4d.utils import GetAngle, Neighbor, VectorEqual, SendModelingCommand
from c4d.plugins import GeLoadString as txt
from math import isnan

MODULE_ID = 1057121
res = ""

VONC_SELECTION_SS =          300001
VONC_SELECTION_SS_VAL =      300101
VONC_SELECTION_SS_FER =      300102
VONC_SELECTION_SELDIR_NOM =  301001
VONC_SELECTION_SELDIR_TOL =  301002
VONC_SELECTION_SELDIR_LIM =  301003
VONC_SELECTION_SELDIR_AID =  301004
VONC_SELECTION_SELDIR_NOR =  301005
VONC_SELECTION_SELNSN_NOM =  301101
VONC_SELECTION_SELNSN_SEL =  301102
VONC_SELECTION_SELNSN_SUR =  301103
VONC_SELECTION_SELNSN_DEC =  301104
VONC_SELECTION_SELNSN_AID =  301105
VONC_SELECTION_SELTAIL_NOM = 301201
VONC_SELECTION_SELTAIL_TOL = 301202
VONC_SELECTION_SELTAIL_INF = 301203
VONC_SELECTION_SELTAIL_SUP = 301204
VONC_SELECTION_SELTAIL_AID = 301205
VONC_SELECTION_SELTRI_NOM =  301301
VONC_SELECTION_SELTRI_AID =  301302
VONC_SELECTION_SELQUAD_NOM = 301401
VONC_SELECTION_SELQUAD_AID = 301402
VONC_SELECTION_SELNGON_NOM = 301501
VONC_SELECTION_SELNGON_AID = 301502
VONC_SELECTION_SELCON_NOM =  301601
VONC_SELECTION_SELCON_AID =  301602
VONC_SELECTION_SELIND_NOM =  301701
VONC_SELECTION_SELIND_AID =  301702
VONC_SELECTION_SELIND_VAL =  301703
VONC_SELECTION_SELIND_DAL =  301704
VONC_SELECTION_SELADJ_NOM =  301801
VONC_SELECTION_SELADJ_AID =  301802
VONC_SELECTION_SELPOS_NOM =  301901
VONC_SELECTION_SELPOS_TOL =  301902
VONC_SELECTION_SELPOS_AID =  301903
VONC_SELECTION_SELPIF_NOM =  302001
VONC_SELECTION_SELPIF_AID =  302002
VONC_SELECTION_SELPIF_SRC =  302003
VONC_SELECTION_SELPIF_QTE =  302004
VONC_SELECTION_SELPIF_CHC =  302005
VONC_SELECTION_SELMIR_NOM =  302101
VONC_SELECTION_SELMIR_AID =  302102
VONC_SELECTION_SELMIR_TOL =  302103
VONC_SELECTION_SELMIR_AJS =  302104
VONC_SELECTION_SELPTNU_NOM = 302201
VONC_SELECTION_SELPTNU_AID = 302202
VONC_SELECTION_SELSURF_NOM = 302301
VONC_SELECTION_SELSURF_AID = 302302
VONC_SELECTION_SELSURF_PREC = 302303
VONC_SELECTION_SELSURF_OBJ = 302304

def icone(nom):
    bmp = bitmaps.BaseBitmap()
    bmp.InitWith(os.path.join(res, nom))
    return bmp

class SelectionAbstract(object) :

    def __init__(self, dialog, group, doc):
        self.dialog = dialog
        self.group = group
        self.doc = doc

    def CreateLayout(self) :
        return True

    def InitValues(self) :
        return True

    def Command(self, id, msg) :
        return True

    def CoreMessage(self, id, msg) :
        return True


class SelectionDirection(SelectionAbstract) :
    tolerance = 0.17453
    limite = False
    norbool = False
    normale = Vector()

    def CreateLayout(self) :
        self.dialog.GroupBegin(100, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)

        self.dialog.GroupBegin(101, c4d.BFH_SCALEFIT, 2, 1)
        self.dialog.AddStaticText(20, c4d.BFH_FIT, 0, 0, txt(VONC_SELECTION_SELDIR_TOL))
        self.dialog.AddEditSlider(11, c4d.BFH_SCALEFIT)
        self.dialog.GroupEnd()

        self.dialog.GroupBegin(102, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.AddCheckbox(12, c4d.BFH_SCALEFIT, 0, 0, txt(VONC_SELECTION_SELDIR_LIM))
        self.dialog.AddCheckbox(13, c4d.BFH_SCALEFIT, 0, 0, txt(VONC_SELECTION_SELDIR_NOR))
        self.dialog.GroupEnd()

        self.dialog.GroupBegin(104, c4d.BFH_SCALEFIT, 3, 1)
        self.dialog.AddEditNumberArrows(130, c4d.BFH_SCALEFIT)
        self.dialog.AddEditNumberArrows(131, c4d.BFH_SCALEFIT)
        self.dialog.AddEditNumberArrows(132, c4d.BFH_SCALEFIT)
        self.dialog.GroupEnd()

        self.dialog.GroupEnd()
        return True

    def InitValues(self) :
        self.dialog.SetDegree(11, self.tolerance, 0, 360)
        self.dialog.SetFloat(130, 0.0, step=1.0)
        self.dialog.SetFloat(131, 0.0, step=1.0)
        self.dialog.SetFloat(132, 0.0, step=1.0)
        self.dialog.Enable(130, False)
        self.dialog.Enable(131, False)
        self.dialog.Enable(132, False)
        return True

    def Command(self, id, msg) :
        if id == 13 :
            nor = self.dialog.GetBool(13)
            self.dialog.Enable(130, nor)
            self.dialog.Enable(131, nor)
            self.dialog.Enable(132, nor)

        if id == c4d.DLG_OK :
            self.tolerance = self.dialog.GetFloat(11)
            self.limite = self.dialog.GetBool(12)
            self.norbool = self.dialog.GetBool(13)
            self.normale = self.dialog.GetVector(130, 131, 132)

            obj = self.doc.GetActiveObject()
            self.sel_dir(self.doc, obj)

        return True


    def recup_norm(self, index, obj) :
        poly = obj.GetPolygon(index)
        pA = obj.GetPoint(poly.a)
        pB = obj.GetPoint(poly.b)
        pC = obj.GetPoint(poly.c)
        pD = obj.GetPoint(poly.d)
        normale = (pA - pC).Cross(pB - pD)
        normale.Normalize()
        return normale

    def verif(self, va, vb, tol) :
        angle = GetAngle(va, vb)
        if isnan(angle) :
            if VectorEqual(va, -vb) : return False
            return True
        if tol >= angle : return True
        else : return False

    def sel_dir(self, doc, obj) :
        if not obj : return
        if obj.GetType() != c4d.Opolygon : return
        if doc.GetMode() != c4d.Mpolygons : return

        bs = obj.GetPolygonS()
        nbsel = bs.GetCount()
        if nbsel == 0 and not self.norbool : return

        nbpol = obj.GetPolygonCount()
        normale_ref = Vector(0, 0, 0)

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        premsel = 0
        for index, selec in enumerate(bs.GetAll(nbpol)) :
            if not selec: continue
            normale = self.recup_norm(index, obj)
            normale_ref += normale
            premsel = index

        if self.norbool : normale_ref = self.normale

        normale_ref.Normalize()
        tolerance = self.tolerance
        limite = self.limite

        for i in range(nbpol) :
            normale = self.recup_norm(i, obj)
            if self.verif(normale_ref, normale, tolerance) :
                bs.Select(i)

        if limite and nbsel :
            commande = c4d.ID_MODELING_FILL_SELECTION_TOOL
            mode = c4d.MODELINGCOMMANDMODE_POLYGONSELECTION
            params = c4d.BaseContainer()

            startPolyID = c4d.MDATA_FILLSEL_START_POLY if c4d.GetC4DVersion() < 21600 else 1100  # MDATA_FILL_POLYGON_INDEX in C++
            selectionTypeID = c4d.MDATA_FILLSEL_SELECTION if c4d.GetC4DVersion() < 21600 else 1110  # MDATA_FILL_SELECTION_TYPE in C++

            params[startPolyID] = premsel
            params[selectionTypeID] = c4d.SELECTION_NEW
            SendModelingCommand(command=commande, list=[obj], mode=mode, bc=params, doc=doc, flags=c4d.MODELINGCOMMANDFLAGS_0)

        doc.EndUndo()
        c4d.EventAdd()
        c4d.CallCommand(12187) # Polygones


class SelectionAlternative(SelectionAbstract) :
    a = 1
    b = 2
    dec = 0

    def CreateLayout(self) :
        self.dialog.GroupBegin(10, c4d.BFH_SCALEFIT, 2, 3)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)
        self.dialog.AddStaticText(20, c4d.BFH_SCALEFIT, 0, 0, txt(VONC_SELECTION_SELNSN_SEL))
        self.dialog.AddEditNumberArrows(30, c4d.BFH_SCALEFIT)
        self.dialog.AddStaticText(21, c4d.BFH_SCALEFIT, 0, 0, txt(VONC_SELECTION_SELNSN_SUR))
        self.dialog.AddEditNumberArrows(31, c4d.BFH_SCALEFIT)
        self.dialog.AddStaticText(22, c4d.BFH_SCALEFIT, 0, 0, txt(VONC_SELECTION_SELNSN_DEC))
        self.dialog.AddEditNumberArrows(32, c4d.BFH_SCALEFIT)
        self.dialog.GroupEnd()
        return True

    def InitValues(self) :
        self.dialog.SetInt32(30, 1, 1)
        self.dialog.SetInt32(31, 2, 1)
        return True

    def Command(self, id, msg) :
        if id == 1 :
            self.a = self.dialog.GetInt32(30)
            self.b = self.dialog.GetInt32(31)
            self.dec = self.dialog.GetInt32(32)
            doc = documents.GetActiveDocument()
            obj = doc.GetActiveObject()
            self.sel_nsn(doc, obj)
        return True

    def verif(self, p, li, bs) :
        if not bs.IsSelected(p) : return False
        if li.IsSelected(p) : return False
        return True

    def point_voisin(self, pp, li, bs, n, obj) :
        if pp is None : return None

        polys = n.GetPointPolys(pp)
        points = []
        for poly in polys :
            poly = obj.GetPolygon(poly)
            points.extend([poly.a, poly.b, poly.c, poly.d])
        points = list(set(points))

        points_aretes = []
        for pt in points :
            arete = n.GetEdgePolys(pp, pt)
            if arete != (-1, -1) : points_aretes.append(pt)

        for pt in points_aretes :
            if not bs.IsSelected(pt) : continue
            if li.IsSelected(pt) : continue
            return pt

        return None

    def mono_voisin(self, pp, li, bs, n, obj) :
        if pp is None : return None
        pol = obj.GetPolygon(pp)

        nab = n.GetNeighbor(pol.a, pol.b, pp)
        if self.verif(nab, li, bs) : return nab

        nbc = n.GetNeighbor(pol.b, pol.c, pp)
        if self.verif(nbc, li, bs) : return nbc

        ncd = n.GetNeighbor(pol.c, pol.d, pp)
        if self.verif(ncd, li, bs) : return ncd

        nda = n.GetNeighbor(pol.d, pol.a, pp)
        if self.verif(nda, li, bs) : return nda

        return None

    def sel_nsn_aretes(self, doc, obj, an, bn, dec) :

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        nbPolys = obj.GetPolygonCount()
        bsAretes = obj.GetEdgeS()
        bs = c4d.BaseSelect()

        n = Neighbor()
        n.Init(obj)

        if an == bn:
            if not bsAretes.GetCount():
                bs.SelectAll(n.GetEdgeCount() - 1)
                obj.SetSelectedEdges(n, bs, c4d.EDGESELECTIONTYPE_SELECTION)
            doc.EndUndo()
            c4d.EventAdd()
            return

        if not bsAretes.GetCount():
            bsAretes.SelectAll(nbPolys * 4 - 1)

        nbPoints = obj.GetPointCount()
        bsTous = bsAretes.GetAll(nbPolys * 4)
        polys = obj.GetAllPolygons()

        pointsLies = [None] * nbPoints
        for i in range(nbPoints):
            pointsLies[i] = set()

        for i, sel in enumerate(bsTous):
            if not sel :
                continue
            ip = int(i / 4)
            poly = polys[ip]
            side = i % 4
            if side == 0:
                a = poly.a
                b = poly.b
            elif side == 1:
                a = poly.b
                b = poly.c
            elif side == 2:
                a = poly.c
                b = poly.d
            else:
                a = poly.d
                b = poly.a

            pointsLies[a].add(b)
            pointsLies[b].add(a)

        for i in range(nbPoints):
            pointsLies[i] = list(pointsLies[i])

        aretesLiees = []

        for i in range(nbPoints):
            if not pointsLies[i]:
                continue

            suite = [i]

            for depart in suite:
                while depart != -1:
                    lies = pointsLies[depart]
                    if not lies:
                        break
                    lie = lies.pop()
                    if depart in pointsLies[lie]:
                        pointsLies[lie].remove(depart)
                        aretesLiees.append((depart, lie))
                    depart = lie
                    if lies:
                        suite.append(lies[0])

        bsAretes.DeselectAll()

        compteur = 0

        if dec > 0:
            aretesLiees =  aretesLiees[dec:] + aretesLiees[:dec]

        for aretes in aretesLiees:

            if compteur < an :
                compteur += 1
            else:
                if compteur == bn-1:
                    compteur = 0
                else:
                    compteur += 1
                continue


            a = aretes[0]
            b = aretes[1]
            first, second = n.GetEdgePolys(a, b)
            trouve = False

            for ipoly in (first, second):

                if ipoly == -1:
                    continue

                poly = polys[ipoly]
                pli = n.GetPolyInfo(ipoly)

                sidesPts = (
                    (poly.a, poly.b),
                    (poly.b, poly.c),
                    (poly.c, poly.d),
                    (poly.d, poly.a),
                )

                for side in range(4):
                    if pli["mark"][side] or (side == 2 and poly.c == poly.d):
                        continue
                    sidePts = sidesPts[side]

                    if (sidePts[0] == a and sidePts[1] == b) or (sidePts[0] == b and sidePts[1] == a):

                        bs.Select(pli["edge"][side])

                        trouve = True
                        break

                if trouve:
                    break

        obj.SetSelectedEdges(n, bs, c4d.EDGESELECTIONTYPE_SELECTION)

        doc.EndUndo()
        c4d.EventAdd()

        return

    def sel_nsn(self, doc, obj) :
        if not obj : return
        polygonal = obj.CheckType(c4d.Opolygon)
        cerce = obj.CheckType(c4d.Ospline)

        modePoints = doc.GetMode() == c4d.Mpoints
        modePolys = doc.GetMode() == c4d.Mpolygons
        modeAretes = doc.GetMode() == c4d.Medges

        a = self.a
        b = self.b
        if self.b < self.a : b = a
        dec = self.dec % b

        if not polygonal and not cerce : return
        if cerce and not modePoints : return
        if modeAretes and polygonal :
            self.sel_nsn_aretes(doc, obj, a, b, dec)
            return
        if not modePoints and not modePolys : return

        if modePolys :
            bs = obj.GetPolygonS()
            nbp = obj.GetPolygonCount()
        else :
            bs = obj.GetPointS()
            nbp = obj.GetPointCount()

        bs_total = bs.GetCount()
        if bs_total and bs_total < 2 : return

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        #-----------
        if cerce :
            if not bs_total : bs.SelectAll(nbp-1)
            li = bs.GetClone()
            bs.DeselectAll()
            j = dec
            bs_sel = li.GetAll(nbp)
            for i, sel in enumerate(bs_sel) :
                if not sel :
                    j = dec
                    continue
                if (j%b) < a : bs.Select(i)
                j += 1

            doc.EndUndo()
            c4d.EventAdd()
            return
        #----------

        n = Neighbor()
        n.Init(obj)
        if not bs_total : bs.SelectAll(nbp)

        def traite(li, p, ordre) :
            if p != None :
                li.Select(p)
                ordre.append(p)

        def tour(pp) :
            ordre1 = []
            voisin = self.mono_voisin
            if modePoints : voisin = self.point_voisin

            pa = voisin(pp, li, bs, n, obj)
            traite(li, pp, ordre1)
            traite(li, pa, ordre1)

            po = voisin(pa, li, bs, n, obj)
            while (po != None) :
                traite(li, po, ordre1)
                po = voisin(po, li, bs, n, obj)

            ordre2 = []
            pb = voisin(pp, li, bs, n, obj)
            if pb != None :
                traite(li, pb, ordre2)
                po = voisin(pb, li, bs, n, obj)
            else : po = None
            while (po != None) :
                traite(li, po, ordre2)
                po = voisin(po, li, bs, n, obj)

            ordre1.reverse()
            ordre1.extend(ordre2)
            return ordre1

        ordre = []

        li = bs.GetClone()
        li.DeselectAll()
        bs_sel = bs.GetAll(nbp)
        for i, sel in enumerate(bs_sel) :
            if not sel : continue
            if li.IsSelected(i) : continue
            ordre.extend(tour(i))

        j = 0
        jmax = len(ordre)
        bs.DeselectAll()

        while (j < jmax) :
            for k in range(a) :
                s = j + k + dec
                if s >= jmax : break
                bs.Select(ordre[s])
            j += b

        doc.EndUndo()
        c4d.EventAdd()


class SelectionTri(SelectionAbstract) :

    def sel_tri(self, doc, obj) :
        if not obj : return
        if obj.GetType() != c4d.Opolygon : return
        ng = obj.GetNgonEdgesCompact()
        bs = obj.GetPolygonS()
        n = -1
        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)
        for p in ng :
            n += 1
            if p != 0: continue
            pol = obj.GetPolygon(n)
            if pol.c == pol.d : bs.Select(n)
        doc.EndUndo()
        c4d.EventAdd()
        c4d.CallCommand(12187) # Polygones

    def Command(self, id, msg) :
        if id == 1 :
            self.Execute(self.doc)
        return True

    def Execute(self, doc) :
        obj = doc.GetActiveObject()
        self.sel_tri(doc, obj)
        return True


class SelectionQuad(SelectionAbstract) :
    def sel_quad(self, doc, obj):
        if not obj: return
        if obj.GetType() != c4d.Opolygon: return
        ng = obj.GetNgonEdgesCompact()
        bs = obj.GetPolygonS()
        n = -1
        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)
        for p in ng:
            n += 1
            if p != 0: continue
            pol = obj.GetPolygon(n)
            if pol.c != pol.d : bs.Select(n)
        doc.EndUndo()
        c4d.EventAdd()
        c4d.CallCommand(12187) # Polygones

    def Command(self, id, msg) :
        if id == 1 :
            self.Execute(self.doc)
        return True

    def Execute(self, doc) :
        obj = doc.GetActiveObject()
        self.sel_quad(doc, obj)
        return True


class SelectionNgon(SelectionAbstract) :
    def sel_ngon(self, doc, obj) :
        if not obj: return
        if obj.GetType() != c4d.Opolygon: return
        ng = obj.GetNgonEdgesCompact()
        bs = obj.GetPolygonS()
        n = 0
        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)
        for p in ng:
            if p != 0: bs.Select(n)
            n += 1
        doc.EndUndo()
        c4d.EventAdd()
        c4d.CallCommand(12187) # Polygones

    def Command(self, id, msg) :
        if id == 1 :
            self.Execute(self.doc)
        return True

    def Execute(self, doc) :
        obj = doc.GetActiveObject()
        self.sel_ngon(doc, obj)
        return True


class SelectionContours(SelectionAbstract) :
    def sel_con(self, doc, obj):
        if not obj: return
        if obj.GetType() != c4d.Opolygon: return
        nb_pol = obj.GetPolygonCount()
        bs = c4d.BaseSelect()
        n = Neighbor()
        n.Init(obj)

        for i in range(nb_pol):
            pol = obj.GetPolygon(i)

            pa = pol.a
            pb = pol.b
            pc = pol.c
            pd = pol.d

            nab = n.GetNeighbor(pa, pb, i)
            nbc = n.GetNeighbor(pb, pc, i)
            ncd = n.GetNeighbor(pc, pd, i)
            nda = n.GetNeighbor(pd, pa, i)

            cotes = n.GetPolyInfo(i)["edge"]

            if nab == -1:
                bs.Select(cotes[0])
            if nbc == -1:
                bs.Select(cotes[1])
            if ncd == -1:
                if pc != pd: bs.Select(cotes[2])
            if nda == -1:
                bs.Select(cotes[3])

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)
        obj.SetSelectedEdges(n, bs, 0)
        doc.EndUndo()
        c4d.EventAdd()
        c4d.CallCommand(16351) # ArĂªtes

    def Command(self, id, msg) :
        if id == 1 :
            self.Execute(self.doc)
        return True

    def Execute(self, doc) :
        obj = doc.GetActiveObject()
        self.sel_con(doc, obj)
        return True


class SelectionIndice(SelectionAbstract) :
    indice = 0
    definir = False

    def CreateLayout(self) :
        self.dialog.GroupBegin(100, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)

        self.dialog.GroupBegin(102, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.AddCheckbox(11, c4d.BFH_SCALEFIT, 0, 0, txt(VONC_SELECTION_SELIND_DAL))
        self.dialog.GroupEnd()

        self.dialog.GroupBegin(101, c4d.BFH_SCALEFIT, 2, 1)
        self.dialog.AddStaticText(20, c4d.BFH_FIT, 0, 0, txt(VONC_SELECTION_SELIND_VAL))
        self.dialog.AddEditNumberArrows(10, c4d.BFH_SCALEFIT)
        self.dialog.GroupEnd()

        self.dialog.GroupEnd()
        return True

    def InitValues(self) :

        indice = self.getIndiceSel(self.doc.GetActiveObject())

        self.dialog.SetInt32(10, indice)
        self.dialog.SetBool(11, False)
        self.dialog.Enable(10, False)
        return True

    def getIndiceSel(self, obj):
        if not obj : return 0
        if obj.GetType() != c4d.Opolygon and obj.GetType() != c4d.Ospline : return 0

        if c4d.IsCommandChecked(12187) and obj.GetType() == c4d.Opolygon : # Polygones
            bs = obj.GetPolygonS()
            nbp = obj.GetPolygonCount()
        else :
            bs = obj.GetPointS()
            nbp = obj.GetPointCount()

        nbsel = bs.GetCount()
        if nbsel == 0 : return 0

        indice = 0
        for i, sel in enumerate(bs.GetAll(nbp)) :
            if sel :
                indice = i
        return indice

    def sel_ind(self, doc, obj):
        if not obj : return
        if obj.GetType() != c4d.Opolygon and obj.GetType() != c4d.Ospline : return

        if c4d.IsCommandChecked(12187) and obj.GetType() == c4d.Opolygon : # Polygones
            bs = obj.GetPolygonS()
            nbp = obj.GetPolygonCount()
        else :
            bs = obj.GetPointS()
            nbp = obj.GetPointCount()

        indice = self.indice
        if not self.definir:
            indice = self.getIndiceSel(obj)

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        for i, sel in enumerate(bs.GetAll(nbp)) :
            bs.Select(i)
            if i == indice :
                break

        doc.EndUndo()
        c4d.EventAdd()

    def Command(self, id, msg) :
        if id == 11 :
            if self.dialog.GetBool(11):
                self.dialog.Enable(10, True)
            else:
                self.dialog.Enable(10, False)
        elif id == 1 :
            self.Execute(self.doc)
        return True

    def Execute(self, doc) :
        obj = doc.GetActiveObject()
        self.indice = self.dialog.GetInt32(10)
        self.definir = self.dialog.GetBool(11)
        self.sel_ind(doc, obj)
        return True

    def CoreMessage(self, id, msg) :

        if id == c4d.EVMSG_CHANGE :
            if not self.dialog.GetBool(11) :
                indice = self.getIndiceSel(self.doc.GetActiveObject())
                self.dialog.SetInt32(10, indice)

        return True


class SelectionTaille(SelectionAbstract) :
    tol = 0
    inf = False
    sup = False

    def CreateLayout(self) :
        self.dialog.GroupBegin(10, c4d.BFH_SCALEFIT, 1, 4)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)
        self.dialog.GroupBegin(15, c4d.BFH_SCALEFIT, 2, 1)
        self.dialog.AddStaticText(16, c4d.BFH_SCALEFIT, name = txt(VONC_SELECTION_SELTAIL_TOL))
        self.dialog.AddEditNumberArrows(11, c4d.BFH_SCALEFIT)
        self.dialog.GroupEnd()
        self.dialog.AddCheckbox(13, c4d.BFH_SCALEFIT, initw = 250, inith = 10, name = txt(VONC_SELECTION_SELTAIL_INF))
        self.dialog.AddCheckbox(14, c4d.BFH_SCALEFIT, initw = 250, inith = 10, name = txt(VONC_SELECTION_SELTAIL_SUP))
        self.dialog.GroupEnd()
        return True

    def InitValues(self) :
        self.dialog.SetPercent(11, 0.005, 0)
        return True

    def Command(self, id, msg) :
        if id == 1 :
            self.tol = self.dialog.GetFloat(11)
            self.inf = self.dialog.GetBool(13)
            self.sup = self.dialog.GetBool(14)
            doc = documents.GetActiveDocument()
            obj = doc.GetActiveObject()
            self.sel_tail(doc, obj)
        return True

    def perim(self, i, obj) :
        poly = obj.GetPolygon(i)
        pA = obj.GetPoint(poly.a)
        pB = obj.GetPoint(poly.b)
        pC = obj.GetPoint(poly.c)
        pD = obj.GetPoint(poly.d)
        # Perimetre
        # per = (pA - pB).GetLengthSquared() + (pB - pC).GetLengthSquared() + (pC - pD).GetLengthSquared() + (pD - pA).GetLengthSquared()

        # Area
        if poly.IsTriangle() :
            return (pA - pB).Cross(pA - pC).GetLength() * 0.5

        ac = pA - pC
        return ((pA - pB).Cross(ac).GetLength() + ac.Cross(pA - pD).GetLength()) * 0.5

    def verif(self, ta, tb, tol, inf, sup) :
        tatol = ta*tol
        if inf :
            if tb <= (ta + tatol) :
                 return True

        if sup :
            if tb >= (ta - tatol) :
                 return True

        if tb <= (ta + tatol) :
            if tb >= (ta - tatol) :
                return True

        return False

    def sel_tail(self, doc, obj):
        if not obj : return
        if obj.GetType() != c4d.Opolygon : return
        if doc.GetMode() != c4d.Mpolygons : return

        bs = obj.GetPolygonS()
        nbsel = bs.GetCount()
        if nbsel == 0 : return

        nbpol = obj.GetPolygonCount()
        taille_ref = 0

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        for index, selec in enumerate(bs.GetAll(nbpol)) :
            if not selec: continue
            taille = self.perim(index, obj)
            taille_ref += taille

        mapt = obj.GetPolygonTranslationMap()
        map = mapt[1]

        ngons = [None] * mapt[0]
        for i in range(mapt[0]) :
            ngons[i] = []

        for p, ngon in enumerate(map) :
            ngons[ngon].append(p)

        for i in range(nbpol) :

            ngonPolys = ngons[map[i]]

            if len(ngonPolys) > 1 :
                taille = 0
                for p in ngonPolys:
                    taille += self.perim(p, obj)
                if self.verif(taille_ref, taille, self.tol, self.inf, self.sup) :
                    for p in ngonPolys:
                        bs.Select(p)

            else:
                taille = self.perim(i, obj)
                if self.verif(taille_ref, taille, self.tol, self.inf, self.sup) :
                    bs.Select(i)

        doc.EndUndo()
        c4d.EventAdd()
        c4d.CallCommand(12187) # Polygones


class SelectionAdjacent(SelectionAbstract) :
    def sel_adj(self, doc, obj):
        if not obj : return
        modePoints = doc.GetMode() == c4d.Mpoints
        modePolys = doc.GetMode() == c4d.Mpolygons
        polygonal = obj.CheckType(c4d.Opolygon)
        cerce = obj.CheckType(c4d.Ospline)
        if not polygonal and not cerce : return
        if not modePoints and not modePolys : return
        if cerce and modePolys : return

        nb_pol = []
        n = None
        if polygonal :
            nb_pol = obj.GetPolygonCount()
            bs = obj.GetPolygonS()
            n = Neighbor()
            n.Init(obj)

        nb_pts = obj.GetPointCount()
        bs_p = obj.GetPointS()

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        if modePolys and polygonal :
            for i, sel in enumerate(bs.GetAll(nb_pol)):
                if not sel: continue
                pol = obj.GetPolygon(i)

                cotes = []
                cotes.extend(n.GetPointPolys(pol.a))
                cotes.extend(n.GetPointPolys(pol.b))
                cotes.extend(n.GetPointPolys(pol.c))
                cotes.extend(n.GetPointPolys(pol.d))

                for p in cotes :
                    bs.Select(p)

        else :
            if polygonal :
                for i, sel in enumerate(bs_p.GetAll(nb_pts)):
                    if not sel : continue
                    polys = n.GetPointPolys(i)
                    for poly in polys :
                        p = obj.GetPolygon(poly)
                        bs_p.Select(p.a)
                        bs_p.Select(p.b)
                        bs_p.Select(p.c)
                        bs_p.Select(p.d)
            else :
                for i, sel in enumerate(bs_p.GetAll(nb_pts)):
                    if not sel : continue
                    iP = i + 1
                    iM = i - 1
                    if iP < nb_pts : bs_p.Select(iP)
                    if iM >= 0 : bs_p.Select(iM)

        doc.EndUndo()
        c4d.EventAdd()

    def Command(self, id, msg) :
        if id == 1 :
            self.Execute(self.doc)
        return True

    def Execute(self, doc) :
        obj = doc.GetActiveObject()
        self.sel_adj(doc, obj)
        return True


class SelectionPosition(SelectionAbstract) :
    tol = 0.001
    pos = Vector()
    Xcon = 0
    Ycon = 0
    Zcon = 0

    def CreateLayout(self) :
        self.dialog.GroupBegin(10, c4d.BFH_SCALEFIT, 1, 3)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)

        self.dialog.GroupBegin(12, c4d.BFH_SCALEFIT, 5, 3)
        self.dialog.AddStaticText(20, c4d.BFH_SCALEFIT, name = "X")
        self.dialog.AddCheckbox(21, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = "<")
        self.dialog.AddCheckbox(22, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = ">")
        self.dialog.AddCheckbox(23, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = "=")
        self.dialog.AddEditNumberArrows(24, c4d.BFH_SCALEFIT, initw = 140)

        self.dialog.AddStaticText(30, c4d.BFH_SCALEFIT, name = "Y")
        self.dialog.AddCheckbox(31, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = "<")
        self.dialog.AddCheckbox(32, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = ">")
        self.dialog.AddCheckbox(33, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = "=")
        self.dialog.AddEditNumberArrows(34, c4d.BFH_SCALEFIT, initw = 140)

        self.dialog.AddStaticText(40, c4d.BFH_SCALEFIT, name = "Z")
        self.dialog.AddCheckbox(41, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = "<")
        self.dialog.AddCheckbox(42, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = ">")
        self.dialog.AddCheckbox(43, c4d.BFH_SCALEFIT, initw = 70, inith = 10, name = "=")
        self.dialog.AddEditNumberArrows(44, c4d.BFH_SCALEFIT, initw = 140)
        self.dialog.GroupEnd()

        self.dialog.GroupBegin(14, c4d.BFH_SCALEFIT, 2, 1)
        self.dialog.AddStaticText(13, c4d.BFH_SCALEFIT, name = txt(VONC_SELECTION_SELPOS_TOL))
        self.dialog.AddEditNumberArrows(11, c4d.BFH_SCALEFIT)
        self.dialog.GroupEnd()

        self.dialog.GroupEnd()
        return True

    def InitValues(self) :
        doc = documents.GetActiveDocument()
        obj = doc.GetActiveObject()
        self.pos = self.sel_pos(doc, obj, True)
        self.dialog.SetMeter(11, self.tol, 0)
        if self.pos is None : return True
        self.dialog.SetMeter(24, self.pos.x)
        self.dialog.SetMeter(34, self.pos.y)
        self.dialog.SetMeter(44, self.pos.z)
        return True

    def conditionneur(self, n) :
        inf = self.dialog.GetBool(n+1)
        sup = self.dialog.GetBool(n+2)
        ega = self.dialog.GetBool(n+3)
        # 0 : aucun, 1 : =, 2 : <, 3 : <=, 4 : >, 5 : >=
        return (inf*2 + sup*4 + ega)

    def Command(self, id, msg) :
        if id == 21 and self.dialog.GetBool(21) : self.dialog.SetBool(22, False)
        if id == 22 and self.dialog.GetBool(22) : self.dialog.SetBool(21, False)
        if id == 31 and self.dialog.GetBool(31) : self.dialog.SetBool(32, False)
        if id == 32 and self.dialog.GetBool(32) : self.dialog.SetBool(31, False)
        if id == 41 and self.dialog.GetBool(41) : self.dialog.SetBool(42, False)
        if id == 42 and self.dialog.GetBool(42) : self.dialog.SetBool(41, False)

        if id == 1 :
            self.tol = self.dialog.GetFloat(11)
            if self.pos is None : return True
            self.pos.x = self.dialog.GetFloat(24)
            self.pos.y = self.dialog.GetFloat(34)
            self.pos.z = self.dialog.GetFloat(44)
            self.Xcon = self.conditionneur(20)
            self.Ycon = self.conditionneur(30)
            self.Zcon = self.conditionneur(40)
            doc = documents.GetActiveDocument()
            obj = doc.GetActiveObject()
            self.sel_pos(doc, obj)
        return True

    def CoreMessage(self, id, msg) :
        if id == c4d.EVMSG_CHANGE :
            self.InitValues()
        return True

    def condition(self, pt, pos, con, tol) :
        if con == 1 :
            if pt < (pos+tol) and pt > (pos-tol) : return True
        elif con == 2 :
            if pt < (pos+tol) : return True
        elif con == 3 :
            if pt <= (pos+tol) : return True
        elif con == 4 :
            if pt > (pos-tol) : return True
        elif con == 5 :
            if pt >= (pos-tol) : return True
        return False

    def centrepoly(self, obj, i) :
        poly = obj.GetPolygon(i)
        centre = Vector()
        centre += obj.GetPoint(poly.a)
        centre += obj.GetPoint(poly.b)
        centre += obj.GetPoint(poly.c)
        if poly.c == poly.d :
            centre /= 3.0
        else :
            centre += obj.GetPoint(poly.d)
            centre *= 0.25
        return centre

    def sel_pos(self, doc, obj, init=False) :
        if not obj : return
        polygonal = obj.CheckType(c4d.Opolygon)
        cerce = obj.CheckType(c4d.Ospline)
        if not polygonal and not cerce : return
        mode = doc.GetMode()
        if mode != c4d.Mpoints and mode != c4d.Mpolygons : return
        if mode == c4d.Mpolygons and cerce : return

        if mode == c4d.Mpoints :
            points = True
            bs = obj.GetPointS()
        else :
            points = False
            bs = obj.GetPolygonS()

        nbsel = bs.GetCount()
        if points : nbp = obj.GetPointCount()
        else : nbp = obj.GetPolygonCount()

        if not init :
            doc.StartUndo()
            doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        centrepoly = self.centrepoly
        nb = 0
        p = Vector()
        pos_min = Vector()
        pos_max = Vector()

        for i, sel in enumerate(bs.GetAll(nbp)) :
            if not sel : continue

            if points : p = obj.GetPoint(i)
            else : p = centrepoly(obj, i)

            if p.x < pos_min.x or not nb : pos_min.x = p.x
            if p.y < pos_min.y or not nb : pos_min.y = p.y
            if p.z < pos_min.z or not nb : pos_min.z = p.z

            if p.x > pos_max.x or not nb : pos_max.x = p.x
            if p.y > pos_max.y or not nb : pos_max.y = p.y
            if p.z > pos_max.z or not nb : pos_max.z = p.z

            nb += 1
            if nb == nbp : break

        pos = (pos_min + pos_max) * 0.5

        if init : return pos

        pos = self.pos
        tol = self.tol
        Xcon = self.Xcon
        Ycon = self.Ycon
        Zcon = self.Zcon
        condition = self.condition
        pt = Vector()

        for i in range(nbp) :
            if points :
                pt = obj.GetPoint(i)
            else :
                pt = centrepoly(obj, i)
            cond = 0
            res = 0
            if Xcon :
                cond += condition(pt.x, pos.x, Xcon, tol)
                res += 1
            if Ycon :
                cond += condition(pt.y, pos.y, Ycon, tol)
                res += 1
            if Zcon :
                cond += condition(pt.z, pos.z, Zcon, tol)
                res += 1
            if cond == res :
                bs.Select(i)

        doc.EndUndo()
        c4d.EventAdd()


class SelectionMiroir(SelectionAbstract) :
    tol = 0.001
    Xcon = True
    Ycon = 0
    Zcon = 0
    ajsel = True

    def CreateLayout(self) :
        self.dialog.GroupBegin(10, c4d.BFH_SCALEFIT, 1, 3)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)

        self.dialog.GroupBegin(12, c4d.BFH_SCALEFIT, 3, 1)
        self.dialog.AddCheckbox(21, c4d.BFH_SCALEFIT, initw = 20, inith = 0, name = "-X")
        self.dialog.AddCheckbox(22, c4d.BFH_SCALEFIT, initw = 20, inith = 0, name = "-Y")
        self.dialog.AddCheckbox(23, c4d.BFH_SCALEFIT, initw = 20, inith = 0, name = "-Z")
        self.dialog.GroupEnd()

        self.dialog.GroupBegin(15, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.AddCheckbox(24, c4d.BFH_SCALEFIT, initw = 0, inith = 0, name = txt(VONC_SELECTION_SELMIR_AJS))
        self.dialog.GroupEnd()

        self.dialog.GroupBegin(14, c4d.BFH_SCALEFIT, 2, 1)
        self.dialog.AddStaticText(13, c4d.BFH_SCALEFIT, name = txt(VONC_SELECTION_SELMIR_TOL))
        self.dialog.AddEditNumberArrows(11, c4d.BFH_SCALEFIT)

        self.dialog.GroupEnd()

        self.dialog.GroupEnd()
        return True

    def InitValues(self) :
        doc = self.doc
        obj = doc.GetActiveObject()
        self.dialog.SetMeter(11, self.tol, 0.0001, step = 0.01)
        self.dialog.SetBool(21, self.Xcon)
        self.dialog.SetBool(22, self.Ycon)
        self.dialog.SetBool(23, self.Zcon)
        self.dialog.SetBool(24, self.ajsel)
        return True

    def Command(self, id, msg) :
        if id == 11 : self.tol = self.dialog.GetFloat(11)
        if id == 21 : self.Xcon = self.dialog.GetBool(21)
        if id == 22 : self.Ycon = self.dialog.GetBool(22)
        if id == 23 : self.Zcon = self.dialog.GetBool(23)
        if id == 24 : self.ajsel = self.dialog.GetBool(24)
        if id == 1 :
            self.tol = self.dialog.GetFloat(11)
            self.Xcon = self.dialog.GetBool(21)
            self.Ycon = self.dialog.GetBool(22)
            self.Zcon = self.dialog.GetBool(23)
            doc = documents.GetActiveDocument()
            obj = doc.GetActiveObject()
            self.sel_mir(doc, obj)
        return True

    def _texte(self, vec, tol) :
        vec *= tol
        vec.x = int(vec.x)
        vec.y = int(vec.y)
        vec.z = int(vec.z)
        p_txt = str(int(vec.x)) + "." + str(int(vec.y)) + "." + str(int(vec.z))
        return p_txt, vec

    def _centrepoly(self, obj, poly) :
        centre = Vector()
        centre += obj.GetPoint(poly.a)
        centre += obj.GetPoint(poly.b)
        centre += obj.GetPoint(poly.c)
        if poly.c == poly.d :
            centre /= 3.0
        else :
            centre += obj.GetPoint(poly.d)
            centre *= 0.25
        return centre

    def _centrearete(self, obj, poly, arete, polycen) :
        centre = Vector()
        aretes = [(poly.a, poly.b), (poly.b, poly.c), (poly.c, poly.d), (poly.d, poly.a)]
        for i in range(2) :
            centre += obj.GetPoint(aretes[arete][i])
        centre = centre * 0.5 + polycen
        return centre

    def sel_mir(self, doc, obj) :
        if not obj : return
        polygonal = obj.CheckType(c4d.Opolygon)
        cerce = obj.CheckType(c4d.Ospline)
        if not polygonal and not cerce : return
        mode = doc.GetMode()
        if cerce and mode != c4d.Mpoints : return

        if mode == c4d.Mpoints :
            bs = obj.GetPointS()
            tab = obj.GetAllPoints()
            nbp = obj.GetPointCount()

        elif mode == c4d.Medges :
            bs = obj.GetEdgeS()
            tab = obj.GetAllPolygons()
            nbp = obj.GetPolygonCount()

        elif mode == c4d.Mpolygons :
            bs = obj.GetPolygonS()
            tab = obj.GetAllPolygons()
            nbp = obj.GetPolygonCount()

        else :
            return

        if not bs or not nbp : return
        nbsel = bs.GetCount()
        if not nbsel : return

        dic = {}
        liste = []
        tol = self.tol
        itol = 1.0 / tol
        Xcon = self.Xcon
        Ycon = self.Ycon
        Zcon = self.Zcon
        Mpolys = c4d.Mpolygons
        Maretes = c4d.Medges
        _centrepoly = self._centrepoly
        _centrearete = self._centrearete
        _texte = self._texte

        for i, p in enumerate(tab) :
            if mode == Maretes :
                polycen = _centrepoly(obj, p)
                for j in range(4) :
                    pt = _centrearete(obj, p, j, polycen)
                    p_txt, vec = _texte(pt, itol)
                    if (p_txt in dic) : dic[p_txt].append(i*4 + j)
                    else : dic[p_txt] = [i*4 + j]
                    liste.append(vec)
            else :
                if mode == Mpolys : pt = _centrepoly(obj, p)
                else : pt = p
                p_txt, vec = _texte(pt, itol)
                if (p_txt in dic) : dic[p_txt].append(i)
                else : dic[p_txt] = [i]
                liste.append(vec)

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        if self.ajsel : bsAct = bs
        else :
            bsAct = bs.GetClone()
            bs.DeselectAll()

        if mode == Maretes : nbp = nbp * 4 + 4

        for i, sel in enumerate(bsAct.GetAll(nbp)) :
            if not sel : continue

            vec = liste[i]
            if Xcon : vec.x *= -1
            if Ycon : vec.y *= -1
            if Zcon : vec.z *= -1
            p_txt = str(int(vec.x)) + "." + str(int(vec.y)) + "." + str(int(vec.z))

            try :
                js = dic[p_txt]
            except KeyError :
                pass
            else :
                for j in js :
                    bs.Select(j)

        if mode == Maretes :
            obj.ValidateEdgeSelection(bs)

        doc.EndUndo()
        c4d.EventAdd()


class SelectionPif(SelectionAbstract) :
    src = 123456
    chc = 0
    bool_src = False
    qte = 10
    bool_qte = False

    def CreateLayout(self) :
        self.dialog.GroupBegin(100, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)

        self.dialog.GroupBegin(101, c4d.BFH_SCALEFIT, 2, 2)
        self.dialog.AddCheckbox(12, c4d.BFH_SCALEFIT, initw = 0, inith = 0, name = txt(VONC_SELECTION_SELPIF_SRC))
        self.dialog.AddEditNumberArrows(11, c4d.BFH_SCALEFIT)
        self.dialog.AddCheckbox(16, c4d.BFH_SCALEFIT, initw = 0, inith = 0, name = txt(VONC_SELECTION_SELPIF_QTE))
        self.dialog.AddEditNumberArrows(15, c4d.BFH_SCALEFIT)
        self.dialog.AddStaticText(13, c4d.BFH_SCALEFIT, name = txt(VONC_SELECTION_SELPIF_CHC))
        self.dialog.AddEditNumberArrows(14, c4d.BFH_SCALEFIT)
        self.dialog.GroupEnd()

        self.dialog.GroupEnd()
        return True

    def InitValues(self) :
        self.dialog.SetInt32(11, self.src)
        self.dialog.SetBool(12, self.bool_src)
        self.dialog.SetInt32(15, self.qte, 0)
        self.dialog.SetBool(16, self.bool_qte)
        self.dialog.SetInt32(14, self.chc)
        self.dialog.Enable(11, self.bool_src)
        self.dialog.Enable(15, self.bool_qte)
        self.dialog.Enable(14, not self.bool_qte)
        self.dialog.Enable(11, self.bool_src)
        return True

    def Command(self, id, msg) :
        if id == 12 : self.dialog.Enable(11, self.dialog.GetBool(12))
        if id == 16 :
            self.dialog.Enable(15, self.dialog.GetBool(16))
            self.dialog.Enable(14, not self.dialog.GetBool(16))

        if id == 1 :
            self.src = self.dialog.GetInt32(11)
            self.bool_src = self.dialog.GetBool(12)
            self.qte = self.dialog.GetInt32(15)
            self.bool_qte = self.dialog.GetBool(16)
            self.chc = self.dialog.GetInt32(14)
            doc = documents.GetActiveDocument()
            obj = doc.GetActiveObject()
            self.sel_pif(doc, obj)
        return True

    def sel_pif(self, doc, obj) :
        if not obj : return
        polygonal = obj.CheckType(c4d.Opolygon)
        cerce = obj.CheckType(c4d.Ospline)
        if not polygonal and not cerce : return
        mode = doc.GetMode()
        if mode != c4d.Mpoints and mode != c4d.Mpolygons : return
        if mode == c4d.Mpolygons and cerce : return

        if mode == c4d.Mpoints :
            points = True
            bs = obj.GetPointS()
        else :
            points = False
            bs = obj.GetPolygonS()

        nbsel = bs.GetCount()
        if points : nbp = obj.GetPointCount()
        else : nbp = obj.GetPolygonCount()

        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)
        src = self.src
        chc = abs(self.chc) + 1
        qte = self.qte
        if qte > nbp : qte = nbp
        bs.DeselectAll()

        if self.bool_src : random.seed(src)

        if self.bool_qte :
            liste = [0] * nbp
            for i in range(qte) : liste[i] = 1
            random.shuffle(liste)
            bs.SetAll(liste)

        else :
            if self.chc > 0 :
                for i in range(nbp) :
                    pif = random.randint(0, chc)
                    if pif : bs.Select(i)
            else :
                for i in range(nbp) :
                    pif = random.randint(0, chc)
                    if not pif : bs.Select(i)

        doc.EndUndo()
        c4d.EventAdd()


class SelectionSurface(SelectionAbstract) :
    precision = 0.01
    autresObjs = True

    def CreateLayout(self) :
        self.dialog.GroupBegin(100, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)

        self.dialog.GroupBegin(101, c4d.BFH_SCALEFIT, 2, 2)
        self.dialog.AddStaticText(13, c4d.BFH_SCALEFIT, name = txt(VONC_SELECTION_SELSURF_PREC))
        self.dialog.AddEditNumberArrows(14, c4d.BFH_SCALEFIT)
        self.dialog.AddCheckbox(12, c4d.BFH_SCALEFIT, initw = 0, inith = 0, name = txt(VONC_SELECTION_SELSURF_OBJ))
        self.dialog.GroupEnd()

        self.dialog.GroupEnd()
        return True

    def InitValues(self) :
        self.dialog.SetBool(12, self.autresObjs)
        self.dialog.SetPercent(14, self.precision, 0)
        return True

    def Command(self, id, msg) :
        if id == 1 :
            self.precision = self.dialog.GetFloat(14)
            self.autresObjs = self.dialog.GetBool(12)
            doc = documents.GetActiveDocument()
            obj = doc.GetActiveObject()
            ops = []
            if not obj:
                objs = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_NONE)
                for o in objs:
                    if not o.CheckType(c4d.Opolygon):
                        continue
                    bs = o.GetPolygonS()
                    if bs.GetCount():
                        ops.append(o)
            else:
                ops = [obj]
            self.sel_surface(doc, ops)
        return True

    def sel_surface_obj(self, doc, obj, airesMinMax):
        if not obj or not obj.CheckType(c4d.Opolygon) : return

        polysGroupe, nbGroupes = self.CalculeGroupesParPoly(obj)
        points = obj.GetAllPoints()
        polys = obj.GetAllPolygons()

        aires = [0.0] * nbGroupes

        for i, groupe in enumerate(polysGroupe):
            aires[groupe] += self.AirePolygone(polys[i], points)

        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)

        bs = obj.GetPolygonS()
        trouve = False

        def estDansAires(aireGroupe, airesMinMax):
            for aireMinMax in airesMinMax :
                if aireMinMax[0] <= aireGroupe <= aireMinMax[1]:
                    return True
            return False

        for groupe, aireGroupe in enumerate(aires):
            if estDansAires(aireGroupe, airesMinMax):
                trouve = True
                for i, groupe2 in enumerate(polysGroupe):
                    if groupe == groupe2:
                        bs.Select(i)
        return trouve

    def sel_surface(self, doc, ops) :

        doc.StartUndo()

        if len(ops) == 1:
            bs = ops[0].GetPolygonS()
            if not bs.GetCount():
                doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, ops[0])
                bs.SelectAll(ops[0].GetPolygonCount() - 1)

        airesMinMax = []

        for op in ops:
            if not op or not op.CheckType(c4d.Opolygon) : return

            bs = op.GetPolygonS()
            nbPolys = op.GetPolygonCount()
            points = op.GetAllPoints()

            polysGroupe, nbGroupes = self.CalculeGroupesParPoly(op)
            aires = [0.0] * nbGroupes
            sel = bs.GetAll(nbPolys)
            for index, selected in enumerate(sel):
                if not selected:
                    continue
                groupe = polysGroupe[index]
                aires[groupe] += self.AirePolygone(op.GetPolygon(index), points)

            for aire in aires:
                if aire > 0.00000001:
                    aireMin = aire - aire * self.precision
                    aireMax = aire + aire * self.precision
                    airesMinMax.append((aireMin, aireMax))

        if not airesMinMax:
            doc.EndUndo()
            return


        objs = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN)
        if not objs or self.autresObjs:
            objs = doc.GetObjects()
            for obj in objs:
                objs.extend(obj.GetChildren())

        for obj in objs:
            trouve = self.sel_surface_obj(doc, obj, airesMinMax)
            if trouve:
                doc.SetActiveObject(obj, c4d.SELECTION_ADD)

        doc.EndUndo()
        c4d.EventAdd()

    @staticmethod
    def AirePolygone(pol, pts) :
        a = pts[pol.a]
        b = pts[pol.b]
        c = pts[pol.c]
        if pol.IsTriangle() :
            return (a - b).Cross(a - c).GetLength() * 0.5
        d = pts[pol.d]
        ac = a - c
        return ((a - b).Cross(ac).GetLength() + ac.Cross(a - d).GetLength()) * 0.5

    @staticmethod
    def CalculeGroupesParPoly(obj):
        if not obj: return None, None

        nbPolys = obj.GetPolygonCount()
        polys = obj.GetAllPolygons()

        polysGroupe = [-1] * nbPolys
        nbGroupes = 0

        n = c4d.utils.Neighbor()
        n.Init(obj)

        for i, g in enumerate(polysGroupe):

            if g != -1: continue

            polysATraiter = [i]

            while polysATraiter:

                polysATraiterSuivant = []

                for p in polysATraiter:
                    polysGroupe[p] = nbGroupes

                for p in polysATraiter:

                    poly = polys[p]

                    voisA = n.GetNeighbor(poly.a, poly.b, p)
                    voisB = n.GetNeighbor(poly.b, poly.c, p)
                    voisD = n.GetNeighbor(poly.d, poly.a, p)
                    voisC = c4d.NOTOK

                    if poly.c != poly.d:
                        voisC = n.GetNeighbor(poly.c, poly.d, p)

                    if voisA != c4d.NOTOK and polysGroupe[voisA] == -1:
                        polysATraiterSuivant.append(voisA)

                    if voisB != c4d.NOTOK and polysGroupe[voisB] == -1:
                        polysATraiterSuivant.append(voisB)

                    if voisC != c4d.NOTOK and polysGroupe[voisC] == -1:
                        polysATraiterSuivant.append(voisC)

                    if voisD != c4d.NOTOK and polysGroupe[voisD] == -1:
                        polysATraiterSuivant.append(voisD)

                polysATraiter = list(set(polysATraiterSuivant))

            nbGroupes += 1

        return polysGroupe, nbGroupes

class SelectionIsoles(SelectionAbstract) :
    def sel_ptnu(self, doc, obj):
        if not obj: return
        if not obj.CheckType(c4d.Opoint) : return
        pols = obj.GetAllPolygons()
        bs = obj.GetPointS()
        doc.StartUndo()
        doc.AddUndo(c4d.UNDOTYPE_CHANGE_SELECTION, obj)
        bs.SelectAll(obj.GetPointCount() - 1)
        for p in pols :
            bs.Deselect(p.a)
            bs.Deselect(p.b)
            bs.Deselect(p.c)
            bs.Deselect(p.d)
        doc.EndUndo()
        c4d.EventAdd()
        c4d.CallCommand(12139) # Points

    def Command(self, id, msg) :
        if id == 1 :
            self.Execute(self.doc)
        return True

    def Execute(self, doc) :
        obj = doc.GetActiveObject()
        self.sel_ptnu(doc, obj)
        return True



class VoncSelectionDialog(gui.GeDialog):

    classe = None
    classes = {
        10001 : {
            'icone' : "Cvoncselection_seldir.tif",
            'icone2' : "Cvoncselection_seldir_sel.tif",
            'classe' : SelectionDirection,
            'titre': VONC_SELECTION_SELDIR_NOM,
            'aide': VONC_SELECTION_SELDIR_AID,
            'options': True,
        },
        10002 : {
            'icone' : "Cvoncselection_selnsn.tif",
            'icone2' : "Cvoncselection_selnsn_sel.tif",
            'classe' : SelectionAlternative,
            'titre': VONC_SELECTION_SELNSN_NOM,
            'aide': VONC_SELECTION_SELNSN_AID,
            'options': True,
        },
        10003 : {
            'icone' : "Cvoncselection_seltri.tif",
            'icone2' : "Cvoncselection_seltri_sel.tif",
            'classe' : SelectionTri,
            'titre': VONC_SELECTION_SELTRI_NOM,
            'aide': VONC_SELECTION_SELTRI_AID,
            'options': False,
        },
        10004 : {
            'icone' : "Cvoncselection_selquad.tif",
            'icone2' : "Cvoncselection_selquad_sel.tif",
            'classe' : SelectionQuad,
            'titre': VONC_SELECTION_SELQUAD_NOM,
            'aide': VONC_SELECTION_SELQUAD_AID,
            'options': False,
        },
        10005 : {
            'icone' : "Cvoncselection_selngon.tif",
            'icone2' : "Cvoncselection_selngon_sel.tif",
            'classe' : SelectionNgon,
            'titre': VONC_SELECTION_SELNGON_NOM,
            'aide': VONC_SELECTION_SELNGON_AID,
            'options': False,
        },
        10006 : {
            'icone' : "Cvoncselection_selcon.tif",
            'icone2' : "Cvoncselection_selcon_sel.tif",
            'classe' : SelectionContours,
            'titre': VONC_SELECTION_SELCON_NOM,
            'aide': VONC_SELECTION_SELCON_AID,
            'options': False,
        },
        10007 : {
            'icone' : "Cvoncselection_selind.tif",
            'icone2' : "Cvoncselection_selind_sel.tif",
            'classe' : SelectionIndice,
            'titre': VONC_SELECTION_SELIND_NOM,
            'aide': VONC_SELECTION_SELIND_AID,
            'options': True,
        },
        10008 : {
            'icone' : "Cvoncselection_seltail.tif",
            'icone2' : "Cvoncselection_seltail_sel.tif",
            'classe' : SelectionTaille,
            'titre': VONC_SELECTION_SELTAIL_NOM,
            'aide': VONC_SELECTION_SELTAIL_AID,
            'options': True,
        },
        10009 : {
            'icone' : "Cvoncselection_seladj.tif",
            'icone2' : "Cvoncselection_seladj_sel.tif",
            'classe' : SelectionAdjacent,
            'titre': VONC_SELECTION_SELADJ_NOM,
            'aide': VONC_SELECTION_SELADJ_AID,
            'options': False,
        },
        10010 : {
            'icone' : "Cvoncselection_selpos.tif",
            'icone2' : "Cvoncselection_selpos_sel.tif",
            'classe' : SelectionPosition,
            'titre': VONC_SELECTION_SELPOS_NOM,
            'aide': VONC_SELECTION_SELPOS_AID,
            'options': True,
        },
        10011 : {
            'icone' : "Cvoncselection_selmir.tif",
            'icone2' : "Cvoncselection_selmir_sel.tif",
            'classe' : SelectionMiroir,
            'titre': VONC_SELECTION_SELMIR_NOM,
            'aide': VONC_SELECTION_SELMIR_AID,
            'options': True,
        },
        10012 : {
            'icone' : "Cvoncselection_selpif.tif",
            'icone2' : "Cvoncselection_selpif_sel.tif",
            'classe' : SelectionPif,
            'titre': VONC_SELECTION_SELPIF_NOM,
            'aide': VONC_SELECTION_SELPIF_AID,
            'options': True,
        },
        10013 : {
            'icone' : "Cvoncselection_selptnu.tif",
            'icone2' : "Cvoncselection_selptnu_sel.tif",
            'classe' : SelectionIsoles,
            'titre': VONC_SELECTION_SELPTNU_NOM,
            'aide': VONC_SELECTION_SELPTNU_AID,
            'options': False,
        },
        10014 : {
            'icone' : "Cvoncselection_selsurf.png",
            'icone2' : "Cvoncselection_selsurf_sel.png",
            'classe' : SelectionSurface,
            'titre': VONC_SELECTION_SELSURF_NOM,
            'aide': VONC_SELECTION_SELSURF_AID,
            'options': True,
        },
    }

    def CreateLayout(soi) :
        soi.GroupBegin(10000, c4d.BFH_SCALEFIT, 7, 0)
        soi.GroupBorderSpace(5, 5, 5, 5)

        for cle in soi.classes:
            obj = soi.classes[cle]
            bc = c4d.BaseContainer()
            bc.SetBool(c4d.BITMAPBUTTON_BUTTON, True)
            bc.SetBool(c4d.BITMAPBUTTON_TOGGLE, True)
            bc[c4d.BITMAPBUTTON_OUTBORDER] = c4d.BORDER_BLACK
            # bc[c4d.BITMAPBUTTON_BACKCOLOR] = c4d.COLOR_BGFOCUS
            bc.SetString(c4d.BITMAPBUTTON_TOOLTIP, "<b>"+txt(obj['titre'])+"</b><br>"+txt(obj['aide']))
            bb = soi.AddCustomGui(cle, c4d.CUSTOMGUI_BITMAPBUTTON, '', c4d.BFH_SCALEFIT, 32, 32, bc)
            bb.SetImage(icone(obj['icone']), True)
            bb.SetImage(icone(obj['icone2']), True, True)
            bb.SetToggleState(False)
            obj['bb'] = bb
            obj['state'] = False

        soi.GroupEnd()

        soi.GroupBegin(10097, c4d.BFH_SCALEFIT, 1, 1)
        soi.AddStaticText(10096, c4d.BFH_SCALEFIT, 0, 0, '', c4d.BORDER_THIN_IN)
        soi.GroupEnd()

        soi.GroupBegin(10099, c4d.BFH_SCALEFIT, 1, 1)
        soi.GroupEnd()

        soi.GroupBegin(10098, c4d.BFH_SCALEFIT, 2, 1)
        soi.GroupBorderSpace(5, 5, 5, 5)
        soi.AddButton(c4d.DLG_OK, c4d.BFH_SCALEFIT, initw = 70, inith = 15, name = txt(VONC_SELECTION_SS_VAL))
        soi.AddButton(c4d.DLG_CANCEL, c4d.BFH_SCALEFIT, initw = 70, inith = 15, name = txt(VONC_SELECTION_SS_FER))
        soi.GroupEnd()

        return True

    def InitValues(soi) :
        if soi.classe is not None :
            return soi.classe.InitValues()
        else:
            soi.setClass(soi.classes[10001], documents.GetActiveDocument())
        return True

    def setClass(soi, obj, doc):
        for cle in soi.classes:
            _obj = soi.classes[cle]
            _obj['state'] = False
            if 'bb' in obj:
                _obj['bb'].SetToggleState(_obj['state'])
        soi.classe = obj['classe'](soi, 10099, doc)
        soi.LayoutFlushGroup(10099)
        soi.classe.CreateLayout()
        soi.classe.InitValues()
        soi.LayoutChanged(10099)
        soi.SetString(10096, txt(obj['titre']))
        obj['state'] = not obj['state']
        if 'bb' in obj:
            obj['bb'].SetToggleState(obj['state'])
        # bc = obj['bb'].GetData()
        # print(bc)
        # bc = c4d.BaseContainer()
        # print(bc)
        # bc[c4d.BITMAPBUTTON_BACKCOLOR] = c4d.COLOR_BG_DARK1
        # obj['bb'].SetData(c4d.BITMAPBUTTON_BACKCOLOR, c4d.COLOR_BGFOCUS)
        # obj['bb'].Activate(False)

    def Command(soi, id, msg) :

        if id in soi.classes:
            obj = soi.classes[id]
            soi.setClass(obj, documents.GetActiveDocument())

        elif  id == c4d.DLG_CANCEL :
            soi.Close()

        elif not soi.classe:
            return True

        else:
            soi.classe.Command(id, msg)

        return True

    def CoreMessage(soi, id, msg):

        if soi.classe is not None :
            return soi.classe.CoreMessage(id, msg)

        return True



class VoncSelection(plugins.CommandData):
    dialog = None

    def Execute(self, doc) :
        if self.dialog is None :
            self.dialog = VoncSelectionDialog()
        return self.dialog.Open(dlgtype=c4d.DLG_TYPE_ASYNC, pluginid=MODULE_ID, defaulth=270, defaultw=290)

    def RestoreLayout(self, sec_ref) :
        if self.dialog is None :
            self.dialog = VoncSelectionDialog()
        return self.dialog.Restore(pluginid=MODULE_ID, secret=sec_ref)

    def ExecuteSubID(self, doc, subid) :

        if self.dialog is None :
            self.dialog = VoncSelectionDialog()

        if subid in self.dialog.classes:
            classe = self.dialog.classes[subid]
            if not classe['options']:
                vraieclasse = classe['classe'](None, 10099, doc)
                vraieclasse.Command(c4d.DLG_OK, None)
            else:
                self.dialog.Open(dlgtype=c4d.DLG_TYPE_ASYNC, pluginid=MODULE_ID, defaulth=270, defaultw=290)
                self.dialog.Command(subid, None)

        return True



if __name__ == "__main__":
    path, fn = os.path.split(__file__)
    bmp = bitmaps.BaseBitmap()
    res = os.path.join(path, "res", "icones")
    bmp.InitWith(os.path.join(res, "Cvoncselection.png"))
    plugins.RegisterCommandPlugin(id=MODULE_ID,
                                  str="#$00"+txt(VONC_SELECTION_SS),
                                  info=0,
                                  help=txt(VONC_SELECTION_SS),
                                  dat=VoncSelection(),
                                  icon=bmp
    )
