# coding=utf-8
import os
import sys
import math
import hashlib
import c4d
import random
import json
import time
from c4d import plugins, utils, bitmaps, gui, Vector, Matrix, CPolygon, documents
from c4d.utils import HPBToMatrix, MatrixScale, Neighbor, SendModelingCommand, Clamp, MatrixRotX
from c4d.plugins import GeLoadString

MODULE_ID = 1057118
res = ""

VONC_SELECTION_SS_VAL = 300101
VONC_SELECTION_SS_FER = 300102

VONC_TOOLS                       = 400000
VONC_TOOLS_AID                   = 400001
VONC_TOOLS_NETTOIEMAT_NOM        = 400010
VONC_TOOLS_NETTOIEMAT_AID        = 400011
VONC_TOOLS_NETTOIEMAT_SUP        = 400012
VONC_TOOLS_NETTOIEMAT_TEXMAT     = 400013
VONC_TOOLS_NETTOIEMAT_TEXPOL     = 400014
VONC_TOOLS_NETTOIEMAT_SELMAT     = 400015
VONC_TOOLS_NORMALISERUV_NOM      = 400020
VONC_TOOLS_NORMALISERUV_AID      = 400021
VONC_TOOLS_AIRVOL_NOM            = 401000
VONC_TOOLS_AIRVOL_ETAT_0         = 401001
VONC_TOOLS_AIRVOL_ETAT_1         = 401002
VONC_TOOLS_AIRVOL_AIRE           = 401003
VONC_TOOLS_AIRVOL_VOLUME         = 401004
VONC_TOOLS_AIRVOL_CALCULER       = 401005
VONC_TOOLS_AIRVOL_POLSEL         = 401006
VONC_TOOLS_AIRVOL_LICENCE        = 401007
VONC_TOOLS_AIRVOL_CLEF           = 401008
VONC_TOOLS_AIRVOL_INFO           = 401009
VONC_TOOLS_AIRVOL_AIDE           = 401010
VONC_TOOLS_AIRVOL_OBJSEL         = 401011
VONC_TOOLS_AIRVOL_SOUSOBJ        = 401012
VONC_TOOLS_AIRVOL_TOTAL          = 401013
VONC_TOOLS_AIRVOL_TOTALE         = 401014
VONC_TOOLS_AIRVOL_MAJ            = 401015

VONC_AIRVOL_UNIT_USFLOZ = 11
VONC_AIRVOL_UNIT_L = 12

def icone(nom):
    bmp = bitmaps.BaseBitmap()
    bmp.InitWith(os.path.join(res, nom))
    return bmp

class ToolAbstract(object) :

    def __init__(self, dialog, group, doc):
        self.dialog = dialog
        self.group = group
        self.doc = doc

    def CreateLayout(self) :
        return True

    def InitValues(self) :
        return True

    def Command(self, id, msg) :
        return True

    def CoreMessage(self, id, msg) :
        return True

class ToolNettoieMat(ToolAbstract):
    vireTexmat = True
    vireTexpol = True
    vireSelmat = True

    def CreateLayout(self) :
        self.dialog.GroupBegin(100, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.GroupBorderSpace(5, 5, 5, 5)

        self.dialog.AddStaticText(20, c4d.BFH_SCALEFIT, 0, 0, GeLoadString(VONC_TOOLS_NETTOIEMAT_SUP))
        self.dialog.AddCheckbox(10,   c4d.BFH_SCALEFIT, 0, 0, GeLoadString(VONC_TOOLS_NETTOIEMAT_TEXMAT))
        self.dialog.AddCheckbox(11,   c4d.BFH_SCALEFIT, 0, 0, GeLoadString(VONC_TOOLS_NETTOIEMAT_TEXPOL))
        self.dialog.AddCheckbox(12,   c4d.BFH_SCALEFIT, 0, 0, GeLoadString(VONC_TOOLS_NETTOIEMAT_SELMAT))

        self.dialog.GroupEnd()
        return True

    def InitValues(self) :
        self.dialog.SetBool(10, self.vireTexmat)
        self.dialog.SetBool(11, self.vireTexpol)
        self.dialog.SetBool(12, self.vireSelmat)
        return True

    def Command(self, id, msg) :
        if id == c4d.DLG_OK :
            self.vireTexmat = self.dialog.GetBool(10)
            self.vireTexpol = self.dialog.GetBool(11)
            self.vireSelmat = self.dialog.GetBool(12)
            self.nettoyeur()
        return True

    def vire(self, p) :
        self.doc.AddUndo(c4d.UNDOTYPE_DELETE, p)
        p.Remove()

    def netTex(self, p) :
        mat = p[c4d.TEXTURETAG_MATERIAL]
        if not mat :
            self.vire(p)
            return True
        return False

    def netSelPol(self, p) :
        bs = p.GetBaseSelect()
        nb = bs.GetCount()
        if not nb :
            if self.vireSelmat and self.vireTexpol : self.vire(p)
            return True
        return False

    def nettoyeur(self) :
        doc = self.doc
        objs = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_0)
        if not objs : return

        doc.StartUndo()

        for obj in objs :
            ps = obj.GetTags()
            dic = {}

            for p in ps :
                if p.CheckType(c4d.Tpolygonselection) :
                    if not self.netSelPol(p) :
                        nom = p[c4d.ID_BASELIST_NAME]
                        if nom in dic :
                            dic[nom].append([p, False])
                        else :
                            dic[nom] = [[p, False]]

                if self.vireTexmat :
                    if p.CheckType(c4d.Ttexture) : self.netTex(p)

            ps = obj.GetTags()
            for p in ps :
                if p.CheckType(c4d.Ttexture) :
                    sel = p[c4d.TEXTURETAG_RESTRICTION]
                    if sel :
                        if not sel in dic :
                            if self.vireTexpol :
                                self.vire(p)
                        else :
                            nbpp = len(dic[sel])
                            for ipp in range(nbpp) :
                                dic[sel][ipp][1] = True

            if self.vireSelmat :
                for ppa in dic.values() :
                    for pp in ppa :
                        if not pp[1] :
                            self.vire(pp[0])

        doc.EndUndo()
        c4d.EventAdd()


class ToolNormaliserUV(ToolAbstract):

    def nb_dec(soi, n) :
        n = abs(n)
        dec = n - int(n)
        nb = str(dec)
        nb = len(nb) - 2
        return nb

    def modif_coor(soi, p, tex, lcX, lcY) :
        p -= Vector(tex[c4d.TEXTURETAG_OFFSETX], tex[c4d.TEXTURETAG_OFFSETY], 0)

        if lcX is True :
            p.x /= tex[c4d.TEXTURETAG_LENGTHX]
        else :
            p.x *= tex[c4d.TEXTURETAG_TILESX]

        if lcY is True :
            p.y /= tex[c4d.TEXTURETAG_LENGTHY]
        else :
            p.y *= tex[c4d.TEXTURETAG_TILESY]

        return p

    def converti(soi, obj) :
        tex = obj.GetTag(c4d.Ttexture)
        if not tex : return

        if tex[c4d.TEXTURETAG_PROJECTION] != c4d.TEXTURETAG_PROJECTION_UVW :
            uvw = obj.GetTag(c4d.Tuvw)
            soi.doc.SetActiveTag(tex, c4d.SELECTION_NEW)
            c4d.CallCommand(12235, 12235)  # Générer des coordonnées UVW
            if uvw :
                uvw.Remove()

        uvw = obj.GetTag(c4d.Tuvw)
        if not uvw : return

        polys = uvw.GetDataCount()
        lcX = True  # Longueur ou Carreaux X
        lcY = True  # Longueur ou Carreaux Y
        if soi.nb_dec(tex[c4d.TEXTURETAG_TILESX]) < soi.nb_dec(tex[c4d.TEXTURETAG_LENGTHX] * 100) :
            lcX = False
        if soi.nb_dec(tex[c4d.TEXTURETAG_TILESY]) < soi.nb_dec(tex[c4d.TEXTURETAG_LENGTHY] * 100) :
            lcY = False

        for i in range(polys) :
            uvwdict = uvw.GetSlow(i)
            a = soi.modif_coor(uvwdict["a"], tex, lcX, lcY)
            b = soi.modif_coor(uvwdict["b"], tex, lcX, lcY)
            c = soi.modif_coor(uvwdict["c"], tex, lcX, lcY)
            d = soi.modif_coor(uvwdict["d"], tex, lcX, lcY)
            uvw.SetSlow(i, a, b, c, d)

        tex[c4d.TEXTURETAG_OFFSETX] = 0
        tex[c4d.TEXTURETAG_OFFSETY] = 0
        tex[c4d.TEXTURETAG_LENGTHX] = 1
        tex[c4d.TEXTURETAG_LENGTHY] = 1
        tex[c4d.TEXTURETAG_TILESX] = 1
        tex[c4d.TEXTURETAG_TILESY] = 1

    def convertir_uvw(soi) :
        doc = soi.doc
        objs = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_0)
        nb = len(objs)
        if nb == 0 : return

        doc.StartUndo()

        for obj in objs :
            doc.AddUndo(c4d.UNDOTYPE_CHANGE, obj)
            soi.converti(obj)

        doc.EndUndo()
        c4d.EventAdd()

    def Command(self, id, msg) :
        if id == 1 :
            self.convertir_uvw()
        return True


class Airevolume(object) :

    def __init__(self, so) :
        self.objs = None
        self.sousObj = so
        self.aire = 0
        self.volume = 0
        self.sel = 0
        self.nbObj = 0
        self.liObjs = []
        self.liVol = []
        self.liAire = []

        self.doc = c4d.documents.GetActiveDocument()
        self.op = self.doc.GetActiveObject()

        if so :
            self.objs = self.doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_0)
        else :
            self.objs = self.doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN)

    def fermeTrous(self, obj, mat) :
        obj.SetMg(mat)
        doc = self.doc
        doc.InsertObject(obj)
        polys = obj.GetAllPolygons()
        n = Neighbor()
        n.Init(obj)
        params = c4d.BaseContainer()
        params[c4d.MDATA_CLOSEHOLE_INDEX] = obj
        params[c4d.MDATA_CLOSEHOLE_TRI] = True
        mode = c4d.MODELINGCOMMANDMODE_EDGESELECTION
        commande = c4d.ID_MODELING_CLOSEHOLE_TOOL

        for i, pol in enumerate(polys) :
            nab = n.GetNeighbor(pol.a, pol.b, i)
            nbc = n.GetNeighbor(pol.b, pol.c, i)
            ncd = n.GetNeighbor(pol.c, pol.d, i)
            nda = n.GetNeighbor(pol.d, pol.a, i)
            cotes = n.GetPolyInfo(i)["edge"]

            if nab == -1 :
                params[c4d.MDATA_CLOSEHOLE_EDGE] = cotes[0]
                SendModelingCommand(command=commande, list=[obj], mode=mode, bc=params, doc=doc)
                n.Init(obj)
            if nbc == -1 :
                params[c4d.MDATA_CLOSEHOLE_EDGE] = cotes[1]
                SendModelingCommand(command=commande, list=[obj], mode=mode, bc=params, doc=doc)
                n.Init(obj)
            if ((ncd == -1) and (pol.c != pol.d)) :
                params[c4d.MDATA_CLOSEHOLE_EDGE] = cotes[2]
                SendModelingCommand(command=commande, list=[obj], mode=mode, bc=params, doc=doc)
                n.Init(obj)
            if nda == -1 :
                params[c4d.MDATA_CLOSEHOLE_EDGE] = cotes[3]
                SendModelingCommand(command=commande, list=[obj], mode=mode, bc=params, doc=doc)
                n.Init(obj)

        obj.Remove()
        c4d.EventAdd()
        return obj

    def volumeTetraedre(self, p1, p2, p3) :
        return p1.Dot(p2.Cross(p3)) / 6.0

    def aireTriangle(self, A, B, C) :
        vAB = A - B
        vAC = A - C
        return vAB.Cross(vAC).GetLength() * 0.5

    def Calcule(self) :
        self.aire = 0
        self.volume = 0
        self.sel = 0
        tab3 = []
        for obj in self.objs :
            conv = SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[obj], doc=self.doc)
            if not conv : continue
            conv = conv[0]
            conv.SetMg(obj.GetMg())
            cache = conv.GetDeformCache() or conv.GetCache() or conv
            cache = cache.GetDeformCache() or cache
            if cache : tab3.append(cache)

        self.objs = tab3

        if self.sousObj :
            tab = []
            tab2 = []
            for obj in self.objs :
                tab2.append(obj)
                sobj = obj.GetDown()
                if sobj : tab.append(sobj)
            for obj in tab :
                while obj :
                    objs = [obj]
                    for o in objs :
                        tab2.append(o)
                        objs.extend(o.GetChildren())
                    obj = obj.GetNext()

            self.objs = tab2

        for obj in self.objs :
            aire, volume = self.CalculeObj(obj)
            self.aire += aire
            self.volume += volume
            if aire or volume :
                self.liObjs.append(obj.GetName())
                self.liAire.append(aire)
                self.liVol.append(volume)

        # self.nbObj = len(self.objs)
        self.nbObj = len(self.liObjs)

    def CalculeObj(self, obj) :
        if not obj : return 0, 0
        doc = self.doc

        # Si cerce
        if obj.CheckType(c4d.Ospline) :
            peau = c4d.BaseObject(c4d.Oloft)
            obj.GetClone().InsertUnder(peau)
            retour = SendModelingCommand(command=c4d.MCOMMAND_CURRENTSTATETOOBJECT, list=[peau], doc=doc)
            if not retour : return 0, 0
            if retour[0].CheckType(c4d.Opolygon) :
                obj = retour[0]
            else :
                obj = retour[0].GetDown()
            if not obj : return 0, 0

        if not obj.CheckType(c4d.Opolygon) :
            return 0, 0

        # Détache la sélection
        if obj.GetBit(c4d.BIT_ACTIVE) :
            if doc.GetMode() == c4d.Mpolygons :
                bs = obj.GetPolygonS()
                self.sel += bs.GetCount()
                retour = SendModelingCommand(command=c4d.MCOMMAND_SPLIT, list=[obj],
                                             mode=c4d.MODELINGCOMMANDMODE_POLYGONSELECTION, doc=doc)
                if not retour : return 0, 0
                if isinstance(retour, list) :
                    obj = retour[0]

        # Optimise
        params = c4d.BaseContainer()
        params[c4d.MDATA_OPTIMIZE_TOLERANCE] = 0.001
        params[c4d.MDATA_OPTIMIZE_POINTS] = False
        params[c4d.MDATA_OPTIMIZE_POLYGONS] = True
        params[c4d.MDATA_OPTIMIZE_UNUSEDPOINTS] = True
        SendModelingCommand(command=c4d.MCOMMAND_OPTIMIZE, list=[obj], bc=params, doc=doc)

        # Ferme les trous
        objVol = self.fermeTrous(obj.GetClone(), obj.GetMg())

        # Aligne les normales
        SendModelingCommand(command=c4d.MCOMMAND_ALIGNNORMALS, list=[obj], doc=doc)
        SendModelingCommand(command=c4d.MCOMMAND_ALIGNNORMALS, list=[objVol], doc=doc)

        # Triangule
        SendModelingCommand(command=c4d.MCOMMAND_TRIANGULATE, list=[obj], doc=doc)
        SendModelingCommand(command=c4d.MCOMMAND_TRIANGULATE, list=[objVol], doc=doc)

        # --

        polys = obj.GetAllPolygons()
        pts = obj.GetAllPoints()
        polysVol = objVol.GetAllPolygons()
        ptsVol = objVol.GetAllPoints()

        ech = Vector(obj.GetMg().v1.GetLength(), obj.GetMg().v2.GetLength(), obj.GetMg().v3.GetLength())
        pts = [pt.__rxor__(ech) for pt in pts]
        ech = Vector(objVol.GetMg().v1.GetLength(), objVol.GetMg().v2.GetLength(), objVol.GetMg().v3.GetLength())
        # ech = objVol.GetAbsScale()
        ptsVol = [pt.__rxor__(ech) for pt in ptsVol]

        aire = 0.0
        for pol in polys :
            aire += self.aireTriangle(pts[pol.a], pts[pol.b], pts[pol.c])

        volume = 0.0
        for pol in polysVol :
            volume += self.volumeTetraedre(ptsVol[pol.a], ptsVol[pol.b], ptsVol[pol.c])

        volume = abs(volume)
        return aire, volume


class ToolAireVolume(ToolAbstract) :

    def __init__(self, dialog, group, doc) :
        super(ToolAireVolume, self).__init__(dialog, group, doc)
        self.airevol = None
        self.uniteFin = 0
        self.sousObj = True
        self.majAuto = True
        self.caches = [[], 0, 0, 0, 0]
        self.airevol = Airevolume(self.sousObj)
        self.airevol.Calcule()

        prefs = c4d.GetWorldContainer()
        unitePref = prefs[c4d.WPREF_UNITS_BASIC]
        self.uniteFin = unitePref

    def unite(self, v, valeur) :
        doc = c4d.documents.GetActiveDocument()

        donneesDoc = doc.GetData()[c4d.DOCUMENT_DOCUNIT]
        ech, uniteDoc = donneesDoc.GetUnitScale()

        uniteFin = self.uniteFin

        ufin = ""
        if uniteFin == c4d.UNIT_NONE :
            ufin = ""
        elif uniteFin == c4d.UNIT_KM :
            ufin = "km"
        elif uniteFin == c4d.UNIT_M :
            ufin = "m"
        elif uniteFin == c4d.UNIT_CM :
            ufin = "cm"
        elif uniteFin == c4d.UNIT_MM :
            ufin = "mm"
        elif uniteFin == c4d.UNIT_UM :
            ufin = "µm"
        elif uniteFin == c4d.UNIT_NM :
            ufin = "nm"
        elif uniteFin == c4d.UNIT_MILE :
            ufin = "mi"
        elif uniteFin == c4d.UNIT_YARD :
            ufin = "yd"
        elif uniteFin == c4d.UNIT_FEET :
            ufin = "ft"
        elif uniteFin == c4d.UNIT_INCH :
            ufin = "in"
        elif uniteFin == VONC_AIRVOL_UNIT_USFLOZ :
            ufin = "US fl oz"
        elif uniteFin == VONC_AIRVOL_UNIT_L :
            ufin = "L"

        valeur *= ech ** v

        fac = 1.0
        if uniteDoc == c4d.UNIT_KM :
            fac *= 10.0 ** (3 * v)
        elif uniteDoc == c4d.UNIT_CM :
            fac *= 10.0 ** (-2 * v)
        elif uniteDoc == c4d.UNIT_MM :
            fac *= 10.0 ** (-3 * v)
        elif uniteDoc == c4d.UNIT_UM :
            fac *= 10.0 ** (-6 * v)
        elif uniteDoc == c4d.UNIT_NM :
            fac *= 10.0 ** (-9 * v)
        elif uniteDoc == c4d.UNIT_MILE :
            fac *= 1609.344 ** v
        elif uniteDoc == c4d.UNIT_YARD :
            fac *= 0.9144 ** v
        elif uniteDoc == c4d.UNIT_FEET :
            fac *= 0.3048 ** v
        elif uniteDoc == c4d.UNIT_INCH :
            fac *= 0.0254 ** v

        if ufin == "km" :
            fac *= 10.0 ** (-3 * v)
        elif ufin == "cm" :
            fac *= 10.0 ** (2 * v)
        elif ufin == "mm" :
            fac *= 10.0 ** (3 * v)
        elif ufin == "µm" :
            fac *= 10.0 ** (6 * v)
        elif ufin == "nm" :
            fac *= 10.0 ** (9 * v)
        elif ufin == "mi" :
            fac /= 1609.344 ** v
        elif ufin == "yd" :
            fac /= 0.9144 ** v
        elif ufin == "ft" :
            fac /= 0.3048 ** v
        elif ufin == "in" :
            fac /= 0.0254 ** v
        elif ufin == "US fl oz" :
            fac *= 33814.0227
        elif ufin == "L" :
            fac *= 10.0 ** v

        valeur *= fac

        if ufin == "US fl oz" or ufin == "L" :
            if v != 3: return ""
            v = ''

        if v == 2 :
            v = "²"
        elif v == 3 :
            v = "³"
        if valeur > 10.0 : valeur = round(valeur, 4)
        retour = str(valeur) + " " + ufin + v
        return retour

    def rapport(self) :
        chn = ""
        chn += GeLoadString(VONC_TOOLS_AIRVOL_POLSEL) + " : " + str(self.airevol.sel) + "\r\n"
        chn += GeLoadString(VONC_TOOLS_AIRVOL_OBJSEL) + " : " + str(self.airevol.nbObj) + "\r\n"
        chn += "\r\n"

        liNb = len(self.airevol.liObjs)

        if liNb > 1 :
            chn += GeLoadString(VONC_TOOLS_AIRVOL_AIRE) + " : " + self.unite(2, self.airevol.aire) + "\r\n"
        else :
            chn += GeLoadString(VONC_TOOLS_AIRVOL_AIRE) + " : \r\n"
        for i in range(liNb) :
            chn += "    " + self.airevol.liObjs[i] + " : " + self.unite(2, self.airevol.liAire[i]) + "\r\n"
        chn += "\r\n"

        if liNb > 1 :
            chn += GeLoadString(VONC_TOOLS_AIRVOL_VOLUME) + " : " + self.unite(3, self.airevol.volume) + "\r\n"
        else :
            chn += GeLoadString(VONC_TOOLS_AIRVOL_VOLUME) + " : \r\n"
        for i in range(liNb) :
            if i : chn += "\r\n"
            chn += "    " + self.airevol.liObjs[i] + " : " + self.unite(3, self.airevol.liVol[i])

        return chn

    def opere(self, force=False) :
        self.sousObj = self.dialog.GetBool(23)
        self.majAuto = self.dialog.GetBool(24)

        doc = c4d.documents.GetActiveDocument()
        if self.sousObj :
            objs = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_0)
        else :
            objs = doc.GetActiveObjects(c4d.GETACTIVEOBJECTFLAGS_CHILDREN)
        caches = [objs, 0, 0, 0, 0]
        if objs :
            tabobjs = objs[:]
            for obj in tabobjs :
                caches[1] += obj.GetDirty(c4d.DIRTY_SELECT)
                caches[2] += obj.GetDirty(c4d.DIRTY_MATRIX)
                caches[3] += obj.GetDirty(c4d.DIRTY_DATA)
                caches[4] += obj.GetDirty(c4d.DIRTY_CHILDREN)
                tabobjs.extend(obj.GetChildren())
        if not force :
            if caches == self.caches :
                self.caches = caches
                return

        self.caches = caches
        self.airevol = Airevolume(self.sousObj)
        self.airevol.Calcule()
        self.uniteFin = self.dialog.GetInt32(41)
        self.dialog.SetString(26, self.rapport())

    def CreateLayout(self) :
        self.dialog.GroupBegin(40, c4d.BFH_SCALEFIT, 2, 1)
        self.dialog.GroupBorderSpace(5, 0, 5, 0)
        # self.dialog.AddButton(21, c4d.BFH_SCALE | c4d.BFH_RIGHT, initw=150, inith=20, name=GeLoadString(VONC_TOOLS_AIRVOL_CALCULER))
        self.dialog.AddComboBox(41, c4d.BFH_SCALE | c4d.BFH_CENTER, initw=60)
        self.dialog.AddChild(41, c4d.UNIT_KM, "km")
        self.dialog.AddChild(41, c4d.UNIT_M, "m")
        self.dialog.AddChild(41, c4d.UNIT_CM, "cm")
        self.dialog.AddChild(41, c4d.UNIT_MM, "mm")
        self.dialog.AddChild(41, c4d.UNIT_UM, "µm")
        self.dialog.AddChild(41, c4d.UNIT_NM, "nm")
        self.dialog.AddChild(41, c4d.UNIT_MILE, "mi")
        self.dialog.AddChild(41, c4d.UNIT_YARD, "yd")
        self.dialog.AddChild(41, c4d.UNIT_FEET, "ft")
        self.dialog.AddChild(41, c4d.UNIT_INCH, "in")
        self.dialog.AddChild(41, VONC_AIRVOL_UNIT_USFLOZ, "US fl oz")
        self.dialog.AddChild(41, VONC_AIRVOL_UNIT_L, "L")
        self.dialog.GroupEnd()

        self.dialog.GroupBegin(20, c4d.BFH_SCALEFIT, 1, 1)
        self.dialog.GroupBorderSpace(5, 0, 5, 0)
        self.dialog.AddCheckbox(23, c4d.BFH_SCALEFIT, 0, 0, GeLoadString(VONC_TOOLS_AIRVOL_SOUSOBJ))
        self.dialog.AddCheckbox(24, c4d.BFH_SCALEFIT, 0, 0, GeLoadString(VONC_TOOLS_AIRVOL_MAJ))
        self.dialog.GroupEnd()

        self.dialog.GroupBegin(25, c4d.BFH_SCALEFIT | c4d.BFV_SCALEFIT, 1, 1)
        self.dialog.GroupBorderSpace(5, 0, 5, 0)
        self.dialog.GroupBorderNoTitle(c4d.BORDER_THIN_IN)
        self.dialog.AddMultiLineEditText(26, c4d.BFH_SCALEFIT | c4d.BFV_SCALEFIT, 0, 150)
        self.dialog.GroupEnd()

        return True

    def InitValues(self) :
        self.dialog.SetTitle(GeLoadString(VONC_TOOLS_AIRVOL_NOM))
        self.dialog.SetInt32(41, self.uniteFin)
        self.dialog.SetBool(23, self.sousObj)
        self.dialog.SetBool(24, self.majAuto)
        self.dialog.SetString(26, self.rapport())
        return True

    def Command(self, id, msg) :
        if id == 21 or id == 23 or id == 24 or id == 1 :
            self.opere(True)
        elif id == 41 :
            self.uniteFin = self.dialog.GetInt32(41)
            self.dialog.SetString(26, self.rapport())

        return True

    def CoreMessage(self, id, msg) :

        if self.majAuto and id == c4d.EVMSG_CHANGE :
            self.opere()

        return True


class VoncToolsDialog(gui.GeDialog):

    classe = None
    classes = {
        10001 : {
            'icone' : "Cvonctools_nettoiemat.png",
            'icone2' : "Cvonctools_nettoiemat_sel.png",
            'classe' : ToolNettoieMat,
            'titre': VONC_TOOLS_NETTOIEMAT_NOM,
            'aide': VONC_TOOLS_NETTOIEMAT_AID,
        },
        10002 : {
            'icone' : "Cvonctools_normaliseruv.png",
            'icone2' : "Cvonctools_normaliseruv_sel.png",
            'classe' : ToolNormaliserUV,
            'titre': VONC_TOOLS_NORMALISERUV_NOM,
            'aide': VONC_TOOLS_NORMALISERUV_AID,
        },
        10003 : {
            'icone' : "Cvonctool_airvolume.tif",
            'icone2' : "Cvonctool_airvolume_sel.tif",
            'classe' : ToolAireVolume,
            'titre': VONC_TOOLS_AIRVOL_NOM,
            'aide': VONC_TOOLS_AIRVOL_AIDE,
        },
    }

    def CreateLayout(soi) :
        soi.SetTitle(GeLoadString(VONC_TOOLS))
        soi.GroupBegin(10000, c4d.BFH_SCALEFIT, 7, 0)
        soi.GroupBorderSpace(5, 5, 5, 5)

        for cle in soi.classes:
            obj = soi.classes[cle]
            bc = c4d.BaseContainer()
            bc.SetBool(c4d.BITMAPBUTTON_BUTTON, True)
            bc.SetBool(c4d.BITMAPBUTTON_TOGGLE, True)
            bc[c4d.BITMAPBUTTON_OUTBORDER] = c4d.BORDER_BLACK
            bc.SetString(c4d.BITMAPBUTTON_TOOLTIP, "<b>"+GeLoadString(obj['titre'])+"</b><br>"+GeLoadString(obj['aide']))
            bb = soi.AddCustomGui(cle, c4d.CUSTOMGUI_BITMAPBUTTON, '', c4d.BFH_SCALEFIT, 32, 32, bc)
            bb.SetImage(icone(obj['icone']), True)
            bb.SetImage(icone(obj['icone2']), True, True)
            bb.SetToggleState(False)
            obj['bb'] = bb
            obj['state'] = False

        soi.GroupEnd()

        soi.GroupBegin(10097, c4d.BFH_SCALEFIT, 1, 1)
        soi.AddStaticText(10096, c4d.BFH_SCALEFIT, 0, 0, '', c4d.BORDER_THIN_IN)
        soi.GroupEnd()

        soi.GroupBegin(10099, c4d.BFH_SCALEFIT, 1, 1)
        soi.GroupEnd()

        soi.GroupBegin(10098, c4d.BFH_SCALEFIT, 2, 1)
        soi.GroupBorderSpace(5, 5, 5, 5)
        soi.AddButton(c4d.DLG_OK, c4d.BFH_SCALEFIT, initw = 70, inith = 15, name = GeLoadString(VONC_SELECTION_SS_VAL))
        soi.AddButton(c4d.DLG_CANCEL, c4d.BFH_SCALEFIT, initw = 70, inith = 15, name = GeLoadString(VONC_SELECTION_SS_FER))
        soi.GroupEnd()

        return True

    def InitValues(soi) :
        if soi.classe is not None :
            return soi.classe.InitValues()
        else:
            soi.setClass(soi.classes[10001], documents.GetActiveDocument())
        return True

    def setClass(soi, obj, doc):
        for cle in soi.classes:
            _obj = soi.classes[cle]
            _obj['state'] = False
            _obj['bb'].SetToggleState(_obj['state'])
        soi.classe = obj['classe'](soi, 10099, doc)
        soi.LayoutFlushGroup(10099)
        soi.classe.CreateLayout()
        soi.classe.InitValues()
        soi.LayoutChanged(10099)
        soi.SetString(10096, GeLoadString(obj['titre']))
        obj['state'] = not obj['state']
        obj['bb'].SetToggleState(obj['state'])

    def Command(soi, id, msg) :

        if id in soi.classes:
            obj = soi.classes[id]
            soi.setClass(obj, documents.GetActiveDocument())

        elif  id == c4d.DLG_CANCEL :
            soi.Close()

        elif not soi.classe:
            return True

        else:
            soi.classe.Command(id, msg)

        return True

    def CoreMessage(soi, id, msg):

        if soi.classe is not None :
            return soi.classe.CoreMessage(id, msg)

        return True

class VoncTools(plugins.CommandData):
    dialog = None

    def Execute(self, doc) :
        if self.dialog is None :
            self.dialog = VoncToolsDialog()
        return self.dialog.Open(dlgtype=c4d.DLG_TYPE_ASYNC, pluginid=MODULE_ID, defaulth=300, defaultw=300)

    def RestoreLayout(self, sec_ref) :
        if self.dialog is None :
            self.dialog = VoncToolsDialog()
        return self.dialog.Restore(pluginid=MODULE_ID, secret=sec_ref)

if __name__ == "__main__":
    path, fn = os.path.split(__file__)
    bmp = bitmaps.BaseBitmap()
    res = os.path.join(path, "res", "icones")
    bmp.InitWith(os.path.join(res, "Ovonctools.png"))
    plugins.RegisterCommandPlugin(id=MODULE_ID,
                                  str="#$01"+GeLoadString(VONC_TOOLS),
                                  info=0,
                                  help=GeLoadString(VONC_TOOLS_AID),
                                  dat=VoncTools(),
                                  icon=bmp
    )
