# coding=utf-8
import os
import sys
import math
import hashlib
import c4d
from c4d import plugins, utils, bitmaps, gui, Vector, Matrix
from c4d.utils import HPBToMatrix, MatrixScale, MixVec, CompareFloatTolerant, Clamp

MODULE_ID = 1037749
VONC_GELATINE_NOM = 500000
VONC_GELATINE_ELASTICITE = 1000
VONC_GELATINE_INTENSITE = 1001
VONC_GELATINE_AMORTIS = 1002
VONC_GELATINE_ATTENUATION = 1003
VONC_GELATINE_FORCE = 1004
VONC_GELATINE_ATTENUATION_INIT = 1006
VONC_GELATINE_RAYON = 1009
VONC_GELATINE_MODE = 1010
VONC_GELATINE_MODE_SRC = 0
VONC_GELATINE_MODE_SEL = 1
VONC_GELATINE_MODE_INF = 2
VONC_GELATINE_INIT = 1011
VONC_GELATINE_INVERSER = 1012
VONC_GELATINE_RAYON_MAX = 1013
VONC_GELATINE_SELECTION = 1014

class Gelatine(plugins.ObjectData) :
	
	PRISES = 2
	
	def __init__(soi) :
		
		soi.pointsCache = {}
		soi.tpsPrec = -1
		soi.ansTps = -1
		
		soi.mode = VONC_GELATINE_MODE_SRC
		soi.rayon = 50.
		soi.rayonMax = 300.
		soi.elasticite = .5
		soi.intensite = 1.
		soi.amortis = 0.
		soi.force = 0.
		soi.inverser = False
		soi.selection = ""
		
		soi.ansRayon = None
		soi.ansRayonMax = None
		soi.ansMode = None
		soi.ansSelection = None
		
		soi.courbe = c4d.SplineData()
		soi.courbe.MakeLinearSplineBezier(2)
		soi.courbe.SetKnot(index = 0, vPos = Vector(0., 1., 0.), lFlagsSettings = 0, vTangentLeft = Vector(-.1, 0.4, 0.), vTangentRight = Vector(.1, -0.4, 0.))
		soi.courbe.SetKnot(index = 1, vPos = Vector(1., 0., 0.), lFlagsSettings = 0, vTangentLeft = Vector(-0.5, 0., 0.), vTangentRight = Vector(0.5, 0., 0.))
	
	def GetHandleCount(op) :
		return Gelatine.PRISES
	
	def GetHandle(self, op, i, info) :
		donnees = op.GetDataInstance()
		if donnees is None : return
		
		if i == 0 :
			info.position = Vector(donnees.GetFloat(VONC_GELATINE_RAYON), 0.0, 0.0)
			info.direction = Vector(1.0, 0.0, 0.0)
			info.type = c4d.HANDLECONSTRAINTTYPE_LINEAR
			
		elif i == 1 :
			info.position = Vector(donnees.GetFloat(VONC_GELATINE_RAYON_MAX), 0.0, 0.0)
			info.direction = Vector(1.0, 0.0, 0.0)
			info.type = c4d.HANDLECONSTRAINTTYPE_LINEAR
	
	def DetectHandle(self, op, bd, x, y, qualifier) :
		if qualifier & c4d.QUALIFIER_CTRL : return c4d.NOTOK
		
		mg = op.GetMg()
		ret = c4d.NOTOK
		
		for i in range(self.GetHandleCount()) :
			info = c4d.HandleInfo()
			self.GetHandle(op, i, info)
			if bd.PointInRange(info.position * mg, x, y) :
				ret = i
				if not qualifier & c4d.QUALIFIER_SHIFT : break
		
		return ret
	
	def MoveHandle(self, op, undo, mouse_pos, hit_id, qualifier, bd) :
		dst = op.GetDataInstance()
		
		info = c4d.HandleInfo()
		val = mouse_pos.x
		self.GetHandle(op, hit_id, info)
		
		if bd is not None :
			mg = op.GetUpMg() * undo.GetMl()
			pos = bd.ProjectPointOnLine(info.position * mg, info.direction ^ mg, mouse_pos.x, mouse_pos.y)
			val = (pos[0] * ~mg) * info.direction
		
		if hit_id == 0 :
			dst.SetFloat(VONC_GELATINE_RAYON, utils.FCut(val, 0.0, float('inf')))
		
		elif hit_id == 1 :
			dst.SetFloat(VONC_GELATINE_RAYON_MAX, utils.FCut(val, 0.0, float('inf')))
			
		return True
	
	@staticmethod
	def DrawCercle(bd, bh, rayon, couleur) :
		m = bh.GetMg()
		m.Scale(rayon)
		bd.SetMatrix_Matrix(None, c4d.Matrix())
		# bd.SetPen(bd.GetObjectColor(bh, op))
		bd.SetPen(couleur)
		bd.DrawCircle(m)
		h = m.v2
		m.v2 = m.v3
		m.v3 = h
		bd.DrawCircle(m)
		h = m.v1
		m.v1 = m.v3
		m.v3 = h
		bd.DrawCircle(m)
	
	def Draw(soi, op, drawpass, bd, bh) :
		donnees = op.GetDataInstance()
		if donnees is None : return c4d.DRAWRESULT_SKIP
		
		mode = donnees.GetInt32(VONC_GELATINE_MODE)
			
		if drawpass == c4d.DRAWPASS_OBJECT :
		
			if mode == VONC_GELATINE_MODE_SRC :
				soi.DrawCercle(bd, bh, donnees.GetFloat(VONC_GELATINE_RAYON), Vector(.1, 1., 0.5))
			
			if mode != VONC_GELATINE_MODE_INF :
				soi.DrawCercle(bd, bh, donnees.GetFloat(VONC_GELATINE_RAYON_MAX), Vector(0.663, 0.6, 1.))
				
			return c4d.DRAWRESULT_OK
		
		elif drawpass == c4d.DRAWPASS_HANDLES :
			i = 0
			bd.SetMatrix_Matrix(None, bh.GetMg())
			
			surbrille = op.GetHighlightHandle(bd)
			
			for i in range(soi.PRISES) :
			
				if mode != VONC_GELATINE_MODE_SRC and i == 0 :
					continue
			
				if mode == VONC_GELATINE_MODE_INF :
					continue
			
				info = c4d.HandleInfo()
				soi.GetHandle(op, i, info)
				if i == surbrille :
					bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_SELECTION_PREVIEW))
				else :
					bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_ACTIVEPOINT))
					
				bd.DrawHandle(info.position, c4d.DRAWHANDLE_BIG, 0)
			
			info = c4d.HandleInfo()
			soi.GetHandle(op, 1, info)
			bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_ACTIVEPOINT))
			bd.DrawLine(info.position, Vector(0), 0)
			
			return c4d.DRAWRESULT_OK
		
		return c4d.DRAWRESULT_SKIP
	
	def Init(soi, op) :
		donnees = op.GetDataInstance()
		donnees.SetFloat(VONC_GELATINE_RAYON, soi.rayon)
		donnees.SetFloat(VONC_GELATINE_RAYON_MAX, soi.rayonMax)
		donnees.SetInt32(VONC_GELATINE_MODE, soi.mode)
		donnees.SetFloat(VONC_GELATINE_INTENSITE, soi.intensite)
		donnees.SetFloat(VONC_GELATINE_ELASTICITE, soi.elasticite)
		donnees.SetFloat(VONC_GELATINE_AMORTIS, soi.amortis)
		donnees.SetData(VONC_GELATINE_ATTENUATION, soi.courbe)
		donnees.SetData(VONC_GELATINE_FORCE, soi.force)
		donnees.SetBool(VONC_GELATINE_INVERSER, soi.inverser)
		donnees.SetString(VONC_GELATINE_SELECTION, soi.selection)
		return True
	
	def Recup(soi, op) :
		donnees = op.GetDataInstance()
		soi.rayon = donnees.GetFloat(VONC_GELATINE_RAYON)
		soi.rayonMax = donnees.GetFloat(VONC_GELATINE_RAYON_MAX)
		soi.mode = donnees.GetInt32(VONC_GELATINE_MODE)
		soi.intensite = donnees.GetFloat(VONC_GELATINE_INTENSITE)
		soi.elasticite = donnees.GetFloat(VONC_GELATINE_ELASTICITE)
		soi.amortis = donnees.GetFloat(VONC_GELATINE_AMORTIS)
		soi.force = donnees.GetFloat(VONC_GELATINE_FORCE)
		soi.courbe = donnees.GetData(VONC_GELATINE_ATTENUATION)
		soi.inverser = donnees.GetBool(VONC_GELATINE_INVERSER)
		soi.selection = donnees.GetString(VONC_GELATINE_SELECTION)
		
		if soi.rayonMax < soi.rayon :
			soi.rayonMax = soi.rayon
		
	def CheckDirty(soi, op, doc) :
		tps = doc.GetTime().GetFrame(doc.GetFps())
		if tps != soi.tpsPrec :
			soi.tpsPrec = tps
			op.SetDirty(c4d.DIRTY_DATA)
			
	def RecupVoisinsPoints(soi, op) :
		
		nbPoints = op.GetPointCount()
		voisins = [None] * nbPoints
		polys = op.GetAllPolygons()
		
		for i in range(nbPoints) :
			voisins[i] = []
		
		for poly in polys :
			if poly.c != poly.d :
				voisins[poly.a].append(poly.b)
				voisins[poly.a].append(poly.c)
				voisins[poly.a].append(poly.d)
				voisins[poly.b].append(poly.a)
				voisins[poly.b].append(poly.c)
				voisins[poly.b].append(poly.d)
				voisins[poly.c].append(poly.a)
				voisins[poly.c].append(poly.b)
				voisins[poly.c].append(poly.d)
				voisins[poly.d].append(poly.a)
				voisins[poly.d].append(poly.b)
				voisins[poly.d].append(poly.c)
			else :
				voisins[poly.a].append(poly.b)
				voisins[poly.a].append(poly.c)
				voisins[poly.b].append(poly.a)
				voisins[poly.b].append(poly.c)
				voisins[poly.c].append(poly.a)
				voisins[poly.c].append(poly.b)
			
		for i in range(nbPoints) :
			voisins[i] = list(set(voisins[i]))
		
		return voisins
	
	def RecupDepart(soi, op, op_mg, mod_mg) :
		departs = []
		
		if soi.mode == VONC_GELATINE_MODE_SRC :
			departs = [mod_mg.off]
		
		elif soi.mode == VONC_GELATINE_MODE_SEL :
			departs = []
			points = op.GetAllPoints()
			if soi.selection :
				for tag in op.GetTags() :
					if tag.CheckType(c4d.Tpointselection) and tag.GetName() == soi.selection :
						bs = tag.GetBaseSelect()
						for i, sel in enumerate(bs.GetAll(op.GetPointCount())) :
							if not sel : continue
							pt = points[i]
							pt = (op_mg).Mul(pt)
							departs.append(pt)
		
		return departs
	
	def CalculeFacteur(soi, op, op_mg, mod_mg) :
	
		nbPoints = op.GetPointCount()
		points = op.GetAllPoints()
		
		if soi.mode == VONC_GELATINE_MODE_INF :
			pointsFacteur = [1.] * nbPoints
			return pointsFacteur
		
		# distanceMax = (op.GetRad() * 2.).GetLength()
		distanceMax = soi.rayonMax - soi.rayon
		
		pointsFacteur = [distanceMax * 10.] * nbPoints
		
		departs = soi.RecupDepart(op, op_mg, mod_mg)
		
		for pointDepart in departs :
			
			for i, point in enumerate(points) :
			
				#
				point = (op_mg).Mul(point)
				#
			
				distance = (pointDepart - point).GetLength()
				# if pointsFacteur[i] > distance :
					# pointsFacteur[i] = distance
				distance -= soi.rayon
				if distance < 0. : distance = 0.
				
				if pointsFacteur[i] > distance :
					pointsFacteur[i] = distance
		
		if distanceMax > 0 :
			for i, fac in enumerate(pointsFacteur) :
				pfac = fac / distanceMax
				if pfac > 1. : pfac = 1.
				pointsFacteur[i] = pfac
		
		return pointsFacteur
	
	@staticmethod
	def ObjID(op) :
		return str(op.GetUniqueIP()) + str(op.GetGUID()) + op.GetName()
		# return str(op.GetUniqueIP()) + str(op.GetGUID())
	
	def Libere(soi, mod) :
		soi.pointsCache = {}
	
	def Initialise(soi, op, op_mg, mod_mg, id) :
	
		pointsElasticite = [Vector()] * op.GetPointCount()
		
		pointsPrecAbs = op.GetAllPoints()
		mg = op.GetMg()
		for i, point in enumerate(pointsPrecAbs) :
			pointsPrecAbs[i] = mg.Mul(point)
		
		pointsFacteur = soi.CalculeFacteur(op, op_mg, mod_mg)
		
		soi.pointsCache[id] = { "precAbs" : pointsPrecAbs,
								"elasticite" : pointsElasticite,
								"facteur" : pointsFacteur,
								"secPrec" : -1. }
	
	def RendUniqueOps(soi, mod) :
		objs = []
		ips = {}
		
		if mod.GetUp() is not None :
			objs = mod.GetUp().GetChildren()
			objs.append(mod.GetUp())
		
		for obj in objs :
			ip = obj.GetUniqueIP()
			
			while (ip in ips) :
				ip += 1
			
			obj.SetUniqueIP(ip)
			ips[ip] = True
			
			# Met à jour la propriété sel de points
			if soi.mode == VONC_GELATINE_MODE_SEL :
				for tag in obj.GetTags() :
					if tag.CheckType(c4d.Tpointselection) and tag.GetName() == soi.selection :
						tag.SetDirty(c4d.DIRTY_DATA)
			
			objs.extend(obj.GetChildren())
	
	def DoitInit(soi, op, id) :
		if not id in soi.pointsCache : return True
		if len(soi.pointsCache[id]["precAbs"]) != op.GetPointCount() : return True
		return False
	
	def DoitLiberer(soi, time, doc) :
		libere = False
		tps = time.GetFrame(doc.GetFps())
		
		# Si on est à l'image 0, libérer la mémoire une fois
		if tps != soi.ansTps and tps == 0 : libere = True
		
		# Si certains paramètres ont changé, libérer la mémoire
		if soi.ansRayon != soi.rayon : libere = True
		if soi.ansRayonMax != soi.rayonMax : libere = True
		if soi.ansMode != soi.mode : libere = True
		if soi.ansSelection != soi.selection : libere = True
		
		soi.ansTps = tps
		soi.ansRayon = soi.rayon
		soi.ansRayonMax = soi.rayonMax
		soi.ansMode = soi.mode
		soi.ansSelection = soi.selection
		
		return libere
	
	def ModifyObject(soi, mod, doc, op, op_mg, mod_mg, lod, flags, thread) :
		
		# Si pas de points, terminer
		if not op.CheckType(c4d.Opoint) or not op.GetPointCount() :
			return True
		
		# Récupère les paramètres
		soi.Recup(mod)
		
		# ID de l'objet en cours de traitement
		id = soi.ObjID(op)
		
		# Variables de temps
		time = doc.GetTime()
		sec = time.Get()
		
		# Libère la mémoire ?
		if soi.DoitLiberer(time, doc) :
			soi.Libere(mod)
			soi.RendUniqueOps(mod)
		
		# Si la mémoire est vide ou si op a changé, initialiser ses tableaux de points
		if soi.DoitInit(op, id) :
			soi.Initialise(op, op_mg, mod_mg, id)
		
		# Bool si le temps est passé ou non
		tpsPasse = sec != soi.pointsCache[id]["secPrec"]
		
		
		#--------------------------------------------------------------------------------#
		
		
		# Points et compagnie
		points = op.GetAllPoints()
		weight = op.CalcVertexmap(mod)
		pointsPrecAbs = soi.pointsCache[id]["precAbs"]
		pointsElasticite = soi.pointsCache[id]["elasticite"]
		pointsFacteur = soi.pointsCache[id]["facteur"]
		op_mg_inv = (~op_mg)
		courbeGetPoint = soi.courbe.GetPoint
		force = soi.force
		amortis = soi.amortis
		inverser = soi.inverser
		intensite = soi.intensite
		elasticite = soi.elasticite
		
		# Boucle principale
		for i, point in enumerate(points) :
		
			pointPrecAbs = pointsPrecAbs[i]
			pointAbs = op_mg.Mul(point)
			
			fac = pointsFacteur[i]
			if weight is not None : fac *= 1. - weight[i]
			if inverser : fac = 1. - fac
			
			facVitesse = courbeGetPoint(fac * elasticite).y
			facAmortissement = courbeGetPoint(fac * amortis).y
			
			pointElasticite = pointsElasticite[i] * (1. - facVitesse)
			pointElasticite += (pointAbs - pointPrecAbs) * Clamp(0., 1., force + facVitesse)
			
			pointNouveauAbsFac = pointElasticite * facAmortissement
			
			#
			# if pointNouveauAbsFac.GetLengthSquared() > 10000. :
				# pointNouveauAbsFac = pointNouveauAbsFac.GetNormalized() * (10000.**.5)
			#
			
			pointPrecAbs = pointsPrecAbs[i] + pointNouveauAbsFac
			
			if tpsPasse :
				pointsElasticite[i] = pointElasticite
				pointsPrecAbs[i] = pointPrecAbs
			
			# Limite max
			# if (soi.pointsCache[id]["precAbs"][i] - pointAbs).GetLengthSquared() > 500000. :
				# soi.pointsCache[id]["precAbs"][i] = (soi.pointsCache[id]["precAbs"][i] - pointAbs).GetNormalized() * (500000.**.5) + pointAbs
			#
			
			# pointNouveau = op_mg_inv.Mul(soi.pointsCache[id]["precAbs"][i])
			pointNouveau = op_mg_inv.Mul(pointPrecAbs)
			
			# Intensité
			pointNouveau = MixVec(point, pointNouveau, intensite)
			
			# Définit le nouveau point
			op.SetPoint(i, pointNouveau)
			
			
		op.Message(c4d.MSG_UPDATE)
		
		
		# Variables précédentes
		soi.pointsCache[id]["secPrec"] = sec
		
		return True
		
	def Message(soi, op, type, data) :
	
		if type == c4d.MSG_MENUPREPARE :
			op.SetDeformMode(True)
			
		elif type == c4d.MSG_DESCRIPTION_COMMAND :
			donnees = op.GetDataInstance()
			id = data["id"][0].id
			
			if id == VONC_GELATINE_ATTENUATION_INIT :
				soi.courbe = c4d.SplineData()
				soi.courbe.MakeLinearSplineBezier(2)
				soi.courbe.SetKnot(index = 0, vPos = Vector(0., 1., 0.), lFlagsSettings = 0, vTangentLeft = Vector(-.1, 0.4, 0.), vTangentRight = Vector(.1, -0.4, 0.))
				soi.courbe.SetKnot(index = 1, vPos = Vector(1., 0., 0.), lFlagsSettings = 0, vTangentLeft = Vector(-0.5, 0., 0.), vTangentRight = Vector(0.5, 0., 0.))
				donnees.SetData(VONC_GELATINE_ATTENUATION, soi.courbe)
			
			elif id == VONC_GELATINE_INIT :
				soi.Libere(op)
				soi.RendUniqueOps(op)
				op.SetDirty(c4d.DIRTY_DATA)
			
		return True
		
	def GetDEnabling(self, node, id, t_data, flags, itemdesc):
		
		donnees = node.GetDataInstance()
		if donnees is None : return True
		
		_id = id[0].id
		
		if _id == VONC_GELATINE_RAYON :
			mode = donnees.GetInt32(VONC_GELATINE_MODE)
			if mode != VONC_GELATINE_MODE_SRC :
				return False
		
		elif _id == VONC_GELATINE_RAYON_MAX :
			mode = donnees.GetInt32(VONC_GELATINE_MODE)
			if mode == VONC_GELATINE_MODE_INF :
				return False
		
		elif _id == VONC_GELATINE_SELECTION :
			mode = donnees.GetInt32(VONC_GELATINE_MODE)
			if mode != VONC_GELATINE_MODE_SEL :
				return False
		
		return True
		

if __name__ == "__main__":
	path, fn = os.path.split(__file__)
	bmp = bitmaps.BaseBitmap()
	bmp.InitWith(os.path.join(path, "res", "icones", "Ovoncgelatine.tif"))
	plugins.RegisterObjectPlugin(id=MODULE_ID, str=c4d.plugins.GeLoadString(VONC_GELATINE_NOM),
								g=Gelatine,
								description="Ovoncgelatine", icon=bmp,
								info=c4d.OBJECT_MODIFIER)
